# File : modbus.py
# Author : Sébastien Deriaz
# License : GPL
"""
Modbus TCP and Modbus RTU implementation
"""

# PDU (Protocol Data Unit) format
#
# Modbus TCP
# ┌────────────────┬─────────────┬────────┬─────────┬───────┐
# │ Transaction ID │ Protocol ID │ Length │ Unit ID │ Data  │
# └────────────────┴─────────────┴────────┴─────────┴───────┘
#       2 bytes        2 bytes     2 bytes   1 byte  N bytes
#
#
# Modbus RTU
# ┌───────────────┬────────┬────────┐
# │ Slave address │  Data  │  CRC   │
# └───────────────┴────────┴────────┘
#      1 byte       N bytes  2 bytes
#
# Modbus ASCII
# ┌────────┬───────────────┬────────┬───────┬─────────┐
# │ HEADER │ Slave address │  Data  │  CRC  │ TRAILER │
# └────────┴───────────────┴────────┴───────┴─────────┘
#   1 byte      1 byte       N bytes 2 bytes   1 byte
#
#
# The PDU is built and parsed by the Modbus class, each data block (SDU)
# is constructed by its corresponding ModbusRequestSDU class. SDUs are parsed
# by ModbusResponseSDU classes

# pylint: disable=too-many-lines

from __future__ import annotations

import struct
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from math import ceil
from types import EllipsisType
from typing import cast

from syndesi.component import AdapterFrame

from ..adapters.adapter import Adapter
from ..adapters.ip import IP
from ..adapters.serialport import SerialPort
from ..adapters.timeout import Timeout
from ..tools.errors import ProtocolError, ProtocolReadError
from .protocol import Protocol, ProtocolFrame

MODBUS_TCP_DEFAULT_PORT = 502

MAX_ADDRESS = 0xFFFF
MIN_ADDRESS = 0x0001

MAX_DISCRETE_INPUTS = (
    0x07B0  # 1968. This is to ensure the total PDU length is 255 at most.
)
# This value has been checked and going up to 1976 seems to work but sticking to the
# spec is safer

# Specification says 125, but write_multiple_registers would exceed the allow number
# of bytes in that case
# MAX_NUMBER_OF_REGISTERS = 123

ExceptionCodesType = dict[int, str]


class Endian(Enum):
    """
    Endian enum
    """

    BIG = "big"
    LITTLE = "little"


endian_symbol = {Endian.BIG: ">", Endian.LITTLE: "<"}


def _dm_to_pdu_address(dm_address: int) -> int:
    """
    Convert Modbus data model address to Modbus PDU address

    - Modbus data model starts at address 1
    - Modbus PDU addresses start at 0

    Modbus data model is the one specified in the devices datasheets

    Parameters
    ----------
    dm_address : int
    """
    if dm_address == 0:
        raise ValueError("Address 0 is not valid in Modbus data model")

    return dm_address - 1


def _pdu_to_dm_address(pdu_address: int) -> int:
    """
    Convert Modbus PDU address to Modbus data model address

    - Modbus data model starts at address 1
    - Modbus PDU addresses start at 0

    Modbus data model is the one specified in the devices datasheets

    Parameters
    ----------
    pdu_address : int
    """
    return pdu_address + 1


def _check_discrete_input_output_count(discrete_inputs: int) -> None:
    if not 1 <= discrete_inputs <= MAX_DISCRETE_INPUTS:
        raise ValueError(
            f"Invalid number of inputs/outputs : {discrete_inputs}, it must be in "
            f"the range [1, {MAX_DISCRETE_INPUTS}]"
        )


def _check_address(address: int, max_address: int | None = None) -> None:
    if not MIN_ADDRESS <= address <= MAX_ADDRESS:
        raise ValueError(
            f"Invalid address : {address}, it must be in the range "
            f"[{MIN_ADDRESS},{MAX_ADDRESS}]"
        )

    if max_address is not None:
        if address > max_address:
            raise ValueError(
                f"Invalid address : {address}, it cannot exceed {max_address}"
                " in the current context"
            )


ENDIAN = endian_symbol[Endian.BIG]

# TCP
# 7 bytes + PDU
# RTU
# 1 byte + PDU + 2 bytes
# Worst case is 255 - 7 bytes = 248 bytes

# Size limitation only apply to Modbus RTU
AVAILABLE_PDU_SIZE = 255 - 3

_ASCII_HEADER = b":"
_PROTOCO_ID = 0
_UNIT_ID = 0
_ASCII_TRAILER = b"\r\n"


class ModbusType(Enum):
    """
    Modbus type

    - TCP : Modbus over TCP
    - RTU : Modbus over serial
    - ASCII : Modbus using text based encoding
    """

    RTU = "RTU"
    ASCII = "ASCII"
    TCP = "TCP"


class FunctionCode(Enum):
    """
    Modbus function codes enum
    """

    # Public function codes 1 to 64
    READ_COILS = 0x01
    READ_DISCRETE_INPUTS = 0x02
    READ_HOLDING_REGISTERS = 0x03
    READ_INPUT_REGISTERS = 0x04
    WRITE_SINGLE_COIL = 0x05
    WRITE_SINGLE_REGISTER = 0x06
    READ_EXCEPTION_STATUS = 0x07  # Serial only
    DIAGNOSTICS = 0x08  # Serial only
    GET_COMM_EVENT_COUNTER = 0x0B  # Serial only
    GET_COMM_EVENT_LOG = 0x0C  # Serial only
    WRITE_MULTIPLE_COILS = 0x0F
    WRITE_MULTIPLE_REGISTERS = 0x10
    REPORT_SERVER_ID = 0x11  # Serial only
    READ_FILE_RECORD = 0x14
    WRITE_FILE_RECORD = 0x15
    MASK_WRITE_REGISTER = 0x16
    READ_WRITE_MULTIPLE_REGISTERS = 0x17
    READ_FIFO_QUEUE = 0x18
    ENCAPSULATED_INTERFACE_TRANSPORT = 0x2B
    # User defined function codes 65 to 72
    # Public function codes 73 to 99
    # User defined function codes 100 to 110
    # Public function codes 111 to 127


class DiagnosticsCode(Enum):
    """
    Modbus Diagnostics codes enum
    """

    RETURN_QUERY_DATA = 0x00
    RESTART_COMMUNICATIONS_OPTION = 0x01
    RETURN_DIAGNOSTIC_REGISTER = 0x02
    CHANGE_ASCII_INPUT_DELIMITER = 0x03
    FORCE_LISTEN_ONLY_MODE = 0x05
    # Reserved 0x05 to 0x09
    CLEAR_COUNTERS_AND_DIAGNOSTIC_REGISTER = 0x0A
    RETURN_BUS_MESSAGE_COUNT = 0x0B
    RETURN_BUS_COMMUNICATION_ERROR_COUNT = 0x0C
    RETURN_BUS_EXCEPTION_ERROR_COUNT = 0x0D
    RETURN_SERVER_MESSAGE_COUNT = 0x0E
    RETURN_SERVER_NO_RESPONSE_COUNT = 0x0F
    RETURN_SERVER_NAK_COUNT = 0x10
    RETURN_SERVER_BUSY_COUNT = 0x11
    RETURN_BUS_CHARACTER_OVERRUN_COUNT = 0x12
    # Reserved 0x13
    CLEAR_OVERRUN_COUNTER_AND_FLAG = 0x14
    # Reserved 0x15 to 0xFF


class EncapsulatedInterfaceTransportSubFunctionCodes(Enum):
    """
    Encapsulated interface transport subfunction codes enum
    """

    CANOPEN_GENERAL_REFERENCE_REQUEST_AND_RESPONSE_PDU = 0x0D
    READ_DEVICE_IDENTIFICATION = 0x0E


class DeviceIndentificationObjects(Enum):
    """
    Device identification objects enum
    """

    VENDOR_NAME = 0x00
    PRODUCT_CODE = 0x01
    MAJOR_MINOR_REVISION = 0x02
    VENDOR_URL = 0x03
    PRODUCT_NAME = 0x04
    MODEL_NAME = 0x05
    USER_APPLICATION_NAME = 0x06
    # Reserved 0x07 to 0x7F
    # Private objects 0x80 to 0xFF


SERIAL_LINE_ONLY_CODES = [
    FunctionCode.DIAGNOSTICS,
    FunctionCode.GET_COMM_EVENT_COUNTER,
    FunctionCode.REPORT_SERVER_ID,
    FunctionCode.READ_EXCEPTION_STATUS,
]


def bool_list_to_bytes(lst: list[bool]) -> bytes:
    """
    Convert a list of bool to bytes, LSB first
    """
    byte_count = ceil(len(lst) / 8)
    result: bytes = sum(2**i * int(v) for i, v in enumerate(lst)).to_bytes(
        byte_count, byteorder="little"
    )
    return result


def bytes_to_bool_list(_bytes: bytes, n: int) -> list[bool]:
    """
    Convert bytes to a list of bool, one per bit, LSB first
    """
    return [c == "1" for c in "".join([f"{x:08b}"[::-1] for x in _bytes])][:n]


class TypeCast(Enum):
    """
    Type of cast when storing values in modbus registers
    """

    INT = "int"
    UINT = "uint"
    FLOAT = "float"
    STRING = "str"
    ARRAY = "array"

    def is_number(self) -> bool:
        """
        Return True if the type is a number
        """
        return self in [TypeCast.INT, TypeCast.UINT, TypeCast.FLOAT]


def struct_format(_type: TypeCast, length: int) -> str:
    """
    Convert typecast+length to python struct character
    """
    struct_characters = {
        (TypeCast.INT, 1): "b",
        (TypeCast.INT, 2): "h",
        (TypeCast.INT, 4): "i",
        (TypeCast.INT, 8): "q",
        (TypeCast.UINT, 1): "B",
        (TypeCast.UINT, 2): "H",
        (TypeCast.UINT, 4): "I",  # or 'L'
        (TypeCast.UINT, 8): "Q",
        (TypeCast.FLOAT, 4): "f",
        (TypeCast.FLOAT, 8): "d",
    }

    if _type in [TypeCast.STRING, TypeCast.ARRAY]:
        return f"{length}s"
    try:
        return struct_characters[_type, length]
    except KeyError:
        pass
    raise ValueError(f"Invalid type cast / length combination : {_type} / {length}")


class ModbusError(Exception):
    """
    Generic modbus exception
    """


def modbus_crc(_bytes: bytes) -> int:
    """Calculate modbus CRC from the given buffer"""
    # TODO : Implement
    return 0


def _raise_if_error(sdu: bytes, exceptions: dict[int, str]) -> None:
    if sdu == b"":
        raise ModbusError("Empty response")
    if sdu[0] & 0x80:
        # There is an error
        code = sdu[1]
        if code not in exceptions:
            raise ModbusError(f"Unexpected modbus error code: {code}")
        raise ModbusError(f"{code:02X} : {exceptions[code]}")


class ModbusSDU:
    """
    Modbus Service Data Unit

    Subclasses of this contain either request or response fields to a modbus
    frame
    """

    def make_sdu(self) -> bytes:
        """Generate a bytes array containing the SDU"""
        raise ProtocolError("make_sdu() is not supported for this SDU")

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        """Generate a ModbusSDU instance from a given sdu buffer"""
        raise ProtocolError("parse_sdu() is not supported for this SDU")

    def exceptions(self) -> dict[int, str]:
        """Return a dictionary of exceptions based on their integer code"""
        return {}

    def _check_for_error(self, sdu: bytes) -> None:
        """Check the given sdu buffer for an exception"""
        _raise_if_error(sdu, self.exceptions())


class ModbusRequestSDU(ModbusSDU):
    """
    Base class for modbus request SDUs (Service Data Unit)
    """

    @abstractmethod
    def function_code(self) -> FunctionCode:
        """Return the function code of this modbus request"""

    @classmethod
    def expected_length(cls, pdu_length: int, modbus_type: ModbusType) -> int:
        """
        Return the length of the modbus SDU based on the length of the PDU and modbus type

        Parameters
        ----------
        pdu_length : int
        modbus_type : ModbusType
        """
        if modbus_type == ModbusType.TCP:
            output = pdu_length + 8
        else:
            output = 1 + pdu_length + 2
            if modbus_type == ModbusType.ASCII:
                # Add header and trailer
                output += len(_ASCII_HEADER) + len(_ASCII_TRAILER)
        return output

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        raise NotImplementedError()


# class ModbusResponseSDU(ModbusSDU):
#     def make_sdu(self) -> bytes:
#         """Generate a bytes array containing the SDU"""
#         raise NotImplementedError("make_sdu() not implemented for this SDU")


class SerialLineOnlySDU(ModbusRequestSDU):
    """
    Marker class for serial line only Modbus function codes
    """

    @abstractmethod
    def exceptions(self) -> dict[int, str]: ...


# Read Coils - 0x01
@dataclass
class ReadCoilsSDU(ModbusRequestSDU):
    """Read coils request."""

    start_address: int
    number_of_coils: int

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        coils: list[bool]

    def function_code(self) -> FunctionCode:
        return FunctionCode.READ_COILS

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid Start or end addresses",
            3: "Invalid quantity of outputs",
            4: "Couldn't read coils",
        }

    def __post_init__(self) -> None:
        _check_discrete_input_output_count(self.number_of_coils)
        _check_address(self.start_address, MAX_ADDRESS - self.number_of_coils + 1)

    def make_sdu(self) -> bytes:
        data = struct.pack(
            ENDIAN + "BHH",
            FunctionCode.READ_COILS.value,
            _dm_to_pdu_address(self.start_address),
            self.number_of_coils,
        )

        return data

    def parse_sdu(self, sdu: bytes) -> Response:
        super()._check_for_error(sdu)
        _, n_bytes = struct.unpack(ENDIAN + "BB", sdu[:2])
        coil_bytes = struct.unpack(ENDIAN + f"{n_bytes}s", sdu[2:])[0]
        coils = bytes_to_bool_list(coil_bytes, self.number_of_coils)
        return self.Response(coils)


# Read discrete inputs - 0x02
@dataclass
class ReadDiscreteInputs(ModbusRequestSDU):
    """Read discrete inputs request."""

    start_address: int
    number_of_inputs: int

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        inputs: list[bool]

    def function_code(self) -> FunctionCode:
        return FunctionCode.READ_DISCRETE_INPUTS

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid Start or end addresses",
            3: "Invalid quantity of inputs",
            4: "Couldn't read inputs",
        }

    def __post_init__(self) -> None:
        _check_discrete_input_output_count(self.number_of_inputs)
        _check_address(self.start_address, MAX_ADDRESS - self.number_of_inputs + 1)

    def make_sdu(self) -> bytes:
        sdu = struct.pack(
            ENDIAN + "BHH",
            self.function_code().value,
            _dm_to_pdu_address(self.start_address),
            self.number_of_inputs,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)

        byte_count = ceil(
            self.number_of_inputs / 8
        )  # pre-calculate the number of returned coil value bytes

        _, _, data = struct.unpack(ENDIAN + f"BB{byte_count}s", sdu)
        inputs = bytes_to_bool_list(data, self.number_of_inputs)

        return self.Response(inputs)


# Read discrete inputs - 0x03
@dataclass
class ReadHoldingRegisters(ModbusRequestSDU):
    """Read holding registers request."""

    start_address: int
    number_of_registers: int

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        registers: list[int]

    def function_code(self) -> FunctionCode:
        return FunctionCode.READ_HOLDING_REGISTERS

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid Start or end addresses",
            3: "Invalid quantity of registers",
            4: "Couldn't read registers",
        }

    def __post_init__(self) -> None:
        _check_address(self.start_address, MAX_ADDRESS - self.number_of_registers + 1)

        max_number_of_registers = (AVAILABLE_PDU_SIZE - 2) // 2

        if not 1 <= self.number_of_registers <= max_number_of_registers:
            raise ValueError(
                f"Invalid number of registers : {self.number_of_registers}"
            )

    def make_sdu(self) -> bytes:
        sdu = struct.pack(
            ENDIAN + "BHH",
            self.function_code().value,
            _dm_to_pdu_address(self.start_address),
            self.number_of_registers,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)

        _, _, registers_data = struct.unpack(
            ENDIAN + f"BB{self.number_of_registers * 2}s", sdu
        )
        registers = list(
            struct.unpack(ENDIAN + "H" * self.number_of_registers, registers_data)
        )

        return self.Response(registers)


# Write Single coil - 0x05
@dataclass
class WriteSingleCoilSDU(ModbusRequestSDU):
    """Write single coil request."""

    address: int
    status: bool

    _ON_VALUE = 0xFF00
    _OFF_VALUE = 0x0000

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

    def function_code(self) -> FunctionCode:
        return FunctionCode.WRITE_SINGLE_COIL

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid address",
            3: "Invalid value",
            4: "Couldn't set coil output",
        }

    def __post_init__(self) -> None:
        _check_address(self.address)

    def make_sdu(self) -> bytes:
        sdu = struct.pack(
            ENDIAN + "BHH",
            FunctionCode.WRITE_SINGLE_COIL.value,
            _dm_to_pdu_address(self.address),
            self._ON_VALUE if self.status else self._OFF_VALUE,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        return self.Response()


@dataclass
class ReadInputRegistersSDU(ModbusRequestSDU):
    """Read input registers request."""

    start_address: int
    number_of_registers: int

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        registers: list[int]

    def function_code(self) -> FunctionCode:
        return FunctionCode.READ_INPUT_REGISTERS

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid Start or end addresses",
            3: "Invalid quantity of registers",
            4: "Couldn't read registers",
        }

    def __post_init__(self) -> None:
        _check_address(self.start_address, MAX_ADDRESS - self.number_of_registers + 1)

        max_number_of_registers = (AVAILABLE_PDU_SIZE - 2) // 2

        if not 1 <= self.number_of_registers <= max_number_of_registers:
            raise ValueError(
                f"Invalid number of registers : {self.number_of_registers}"
            )

    def make_sdu(self) -> bytes:
        sdu = struct.pack(
            ENDIAN + "BHH",
            self.function_code().value,
            _dm_to_pdu_address(self.start_address),
            self.number_of_registers,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)

        _, _, registers_data = struct.unpack(
            ENDIAN + f"BB{self.number_of_registers * 2}s", sdu
        )
        registers = list(
            struct.unpack(ENDIAN + "H" * self.number_of_registers, registers_data)
        )

        return self.Response(registers)


@dataclass
class WriteSingleRegisterSDU(ModbusRequestSDU):
    """Write single register request."""

    address: int
    value: int

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

    def function_code(self) -> FunctionCode:
        return FunctionCode.WRITE_SINGLE_REGISTER

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid address",
            3: "Invalid register value",
            4: "Couldn't write register",
        }

    def __post_init__(self) -> None:
        _check_address(self.address)
        if not 0 <= self.value <= 0xFFFF:
            raise ValueError(f"Invalid register value : {self.value}")

    def make_sdu(self) -> bytes:
        sdu = struct.pack(
            ENDIAN + "BHH",
            self.function_code().value,
            _dm_to_pdu_address(self.address),
            self.value,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        if sdu != self.make_sdu():
            raise ProtocolReadError(
                f"Response ({sdu!r}) should match query ({self.make_sdu()!r})"
            )
        return self.Response()


@dataclass
class ReadExceptionStatusSDU(SerialLineOnlySDU):
    """Read exception status request (serial only)."""

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        status: int

    def function_code(self) -> FunctionCode:
        return FunctionCode.READ_EXCEPTION_STATUS

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            4: "Couldn't read exception status",
        }

    def make_sdu(self) -> bytes:
        return struct.pack(ENDIAN + "B", self.function_code().value)

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        _, status = cast(tuple[int, int], struct.unpack(ENDIAN + "BB", sdu))
        return self.Response(status)


@dataclass
class DiagnosticsSDU(SerialLineOnlySDU):
    """Diagnostics request (serial only)."""

    code: DiagnosticsCode
    subfunction_data: bytes
    return_subfunction_bytes: int
    check_response: bool = True

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        data: bytes

    def function_code(self) -> FunctionCode:
        return FunctionCode.DIAGNOSTICS

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Unsuported function code or sub-function code",
            3: "Invalid data value",
            4: "Diagnostic error",
        }

    def make_sdu(self) -> bytes:
        return (
            struct.pack(ENDIAN + "BH", self.function_code().value, self.code.value)
            + self.subfunction_data
        )

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)

        if self.return_subfunction_bytes == 0:
            returned_function, returned_subfunction_integer = struct.unpack(
                ENDIAN + "BH", sdu
            )
            subfunction_returned_data = b""
        else:
            (
                returned_function,
                returned_subfunction_integer,
                subfunction_returned_data,
            ) = cast(
                tuple[int, int, bytes],
                struct.unpack(ENDIAN + f"BH{self.return_subfunction_bytes}s", sdu),
            )

        if self.check_response:
            returned_subfunction = DiagnosticsCode(returned_subfunction_integer)

            if returned_function != self.function_code().value:
                raise ProtocolReadError(
                    f"Invalid returned function code : {returned_function}"
                )
            if returned_subfunction != self.code:
                raise ProtocolReadError(
                    f"Invalid returned subfunction code : {returned_subfunction}"
                )

        return self.Response(subfunction_returned_data)


@dataclass
class GetCommEventCounterSDU(SerialLineOnlySDU):
    """Get communication event counter request (serial only)."""

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        status: int
        event_count: int

    def function_code(self) -> FunctionCode:
        return FunctionCode.GET_COMM_EVENT_COUNTER

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            4: "Couldn't get comm event counter",
        }

    def make_sdu(self) -> bytes:
        return struct.pack(ENDIAN + "B", self.function_code().value)

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        _, status, event_count = struct.unpack(ENDIAN + "BHH", sdu)
        return self.Response(status, event_count)


@dataclass
class GetCommEventLogSDU(SerialLineOnlySDU):
    """Get communication event log request (serial only)."""

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        status: int
        event_count: int
        message_count: int
        events: bytes

    def function_code(self) -> FunctionCode:
        return FunctionCode.GET_COMM_EVENT_LOG

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            4: "Couldn't get comm event log",
        }

    def make_sdu(self) -> bytes:
        return struct.pack(ENDIAN + "B", self.function_code().value)

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        _, byte_count = struct.unpack(ENDIAN + "BB", sdu[:2])
        status, event_count, message_count = struct.unpack(ENDIAN + "HHH", sdu[2:8])
        events = sdu[8 : 8 + (byte_count - 6)]
        return self.Response(status, event_count, message_count, events)


@dataclass
class WriteMultipleCoilsSDU(ModbusRequestSDU):
    """Write multiple coils request."""

    start_address: int
    values: list[bool]

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

    def function_code(self) -> FunctionCode:
        return FunctionCode.WRITE_MULTIPLE_COILS

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid start and/or end addresses",
            3: "Invalid number of outputs and/or byte count",
            4: "Couldn't write outputs",
        }

    def __post_init__(self) -> None:
        number_of_coils = len(self.values)
        _check_discrete_input_output_count(number_of_coils)
        _check_address(self.start_address, MAX_ADDRESS - number_of_coils + 1)

    def make_sdu(self) -> bytes:
        number_of_coils = len(self.values)
        byte_count = ceil(number_of_coils / 8)
        sdu = struct.pack(
            ENDIAN + f"BHHB{byte_count}s",
            self.function_code().value,
            _dm_to_pdu_address(self.start_address),
            number_of_coils,
            byte_count,
            bool_list_to_bytes(self.values),
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)

        _, start_address, coils_written = struct.unpack(ENDIAN + "BHH", sdu)
        if coils_written != len(self.values):
            raise ProtocolError(
                f"Number of coils written ({coils_written}) doesn't match expected "
                f"value : {len(self.values)}"
            )
        if start_address != _dm_to_pdu_address(self.start_address):
            raise ProtocolReadError(
                f"Start address mismatch : {start_address} != "
                "{_dm_to_pdu_address(self.start_address)}"
            )
        return self.Response()


@dataclass
class WriteMultipleRegistersSDU(ModbusRequestSDU):
    """Write multiple registers request."""

    start_address: int
    values: list[int]

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

    def function_code(self) -> FunctionCode:
        return FunctionCode.WRITE_MULTIPLE_REGISTERS

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid start and/or end addresses",
            3: "Invalid number of outputs and/or byte count",
            4: "Couldn't write outputs",
        }

    def __post_init__(self) -> None:
        max_number_of_registers = (AVAILABLE_PDU_SIZE - 6) // 2

        if len(self.values) == 0:
            raise ValueError("Empty register list")

        if len(self.values) > max_number_of_registers:
            raise ValueError(
                f"Cannot set more than {max_number_of_registers} registers at a time"
            )

        _check_address(self.start_address, MAX_ADDRESS - len(self.values) + 1)

    def make_sdu(self) -> bytes:
        byte_count = 2 * len(self.values)
        sdu = struct.pack(
            ENDIAN + f"BHHB{byte_count // 2}H",
            self.function_code().value,
            _dm_to_pdu_address(self.start_address),
            byte_count // 2,
            byte_count,
            *self.values,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)

        _, start_address, registers_written = struct.unpack(ENDIAN + "BHH", sdu)
        if registers_written != len(self.values):
            raise ProtocolError(
                f"Number of registers written ({registers_written}) doesn't match expected "
                f"value : {len(self.values)}"
            )
        if start_address != _dm_to_pdu_address(self.start_address):
            raise ProtocolReadError(
                f"Start address mismatch : {start_address} != "
                "{_dm_to_pdu_address(self.start_address)}"
            )
        return self.Response()


@dataclass
class ReportServerIdSDU(SerialLineOnlySDU):
    """Report server ID request (serial only)."""

    server_id_length: int
    additional_data_length: int

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        server_id: bytes
        run_indicator_status: bool
        additional_data: bytes

    def function_code(self) -> FunctionCode:
        return FunctionCode.REPORT_SERVER_ID

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            4: "Couldn't report slave ID",
        }

    def make_sdu(self) -> bytes:
        return struct.pack(ENDIAN + "B", self.function_code().value)

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        _, byte_count = struct.unpack(ENDIAN + "BB", sdu[:2])
        expected = self.server_id_length + 1 + self.additional_data_length
        if byte_count != expected:
            raise ProtocolReadError(
                f"Invalid byte count : {byte_count}, expected {expected}"
            )
        start = 2
        end = start + self.server_id_length
        server_id = sdu[start:end]
        run_indicator_status = sdu[end] == 0xFF
        additional_data = sdu[end + 1 : end + 1 + self.additional_data_length]
        return self.Response(server_id, run_indicator_status, additional_data)


@dataclass
class ReadFileRecordSDU(ModbusRequestSDU):
    """Read file record request."""

    records: list[tuple[int, int, int]]

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        records_data: list[bytes]

    def function_code(self) -> FunctionCode:
        return FunctionCode.READ_FILE_RECORD

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid parameters",
            3: "Invalid byte count",
            4: "Couldn't read records",
        }

    def __post_init__(self) -> None:
        size_limit = 253
        query_size = 2 + 7 * len(self.records)
        response_size = 2 + len(self.records) + sum(2 * r[2] for r in self.records)
        if query_size > size_limit:
            raise ValueError(f"Number of records is too high : {len(self.records)}")
        if response_size > size_limit:
            raise ValueError(
                f"Sum of records lenghts is too high : {sum(r[2] for r in self.records)}"
            )

    def make_sdu(self) -> bytes:
        reference_type = 6
        sub_req_buffer = b""
        for file_number, record_number, record_length in self.records:
            sub_req_buffer += struct.pack(
                ENDIAN + "BHHH",
                reference_type,
                file_number,
                record_number,
                record_length,
            )

        byte_count = len(sub_req_buffer)
        sdu = struct.pack(
            ENDIAN + f"BB{byte_count}s",
            self.function_code().value,
            byte_count,
            sub_req_buffer,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        _, byte_count = struct.unpack(ENDIAN + "BB", sdu[:2])
        records_data: list[bytes] = []
        offset = 2
        end = 2 + byte_count
        while offset < end:
            length = sdu[offset]
            reference_type = sdu[offset + 1]
            if reference_type != 6:
                raise ProtocolReadError(f"Invalid reference type : {reference_type}")
            data_start = offset + 2
            data_end = data_start + length - 1
            records_data.append(sdu[data_start:data_end])
            offset = data_end

        return self.Response(records_data)


@dataclass
class WriteFileRecordSDU(ModbusRequestSDU):
    """Write file record request."""

    records: list[tuple[int, int, bytes]]

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

    def function_code(self) -> FunctionCode:
        return FunctionCode.WRITE_FILE_RECORD

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid parameters",
            3: "Invalid byte count",
            4: "Couldn't write records",
        }

    def __post_init__(self) -> None:
        if isinstance(self.records, tuple):
            self.records = [self.records]
        elif not isinstance(self.records, list):
            raise TypeError(f"Invalid records type : {self.records}")

    def make_sdu(self) -> bytes:
        reference_type = 6
        sub_req_buffer = b""

        for file_number, record_number, data in self.records:
            sub_req_buffer += struct.pack(
                ENDIAN + f"BHHH{len(data)}s",
                reference_type,
                file_number,
                record_number,
                len(data) // 2,
                data,
            )

        sdu = struct.pack(
            ENDIAN + f"BB{len(sub_req_buffer)}s",
            self.function_code().value,
            len(sub_req_buffer),
            sub_req_buffer,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        if sdu != self.make_sdu():
            raise ProtocolReadError("Response different from query")
        return self.Response()


@dataclass
class MaskWriteRegisterSDU(ModbusRequestSDU):
    """Mask write register request."""

    address: int
    and_mask: int
    or_mask: int

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

    def function_code(self) -> FunctionCode:
        return FunctionCode.MASK_WRITE_REGISTER

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid register address",
            3: "Invalid AND/OR mask",
            4: "Couldn't write register",
        }

    def __post_init__(self) -> None:
        _check_address(self.address)

    def make_sdu(self) -> bytes:
        sdu = struct.pack(
            ENDIAN + "BHHH",
            self.function_code().value,
            _dm_to_pdu_address(self.address),
            self.and_mask,
            self.or_mask,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        if sdu != self.make_sdu():
            raise ProtocolReadError(
                f"Response ({sdu!r}) should match query ({self.make_sdu()!r})"
            )
        return self.Response()


@dataclass
class ReadWriteMultipleRegistersSDU(ModbusRequestSDU):
    """Read/write multiple registers request."""

    read_starting_address: int
    number_of_read_registers: int
    write_starting_address: int
    write_values: list[int]

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        read_values: list[int]

    def function_code(self) -> FunctionCode:
        return FunctionCode.READ_WRITE_MULTIPLE_REGISTERS

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid read/write start/end address",
            3: "Invalid quantity of read/write and/or byte count",
            4: "Couldn't read and/or write registers",
        }

    def __post_init__(self) -> None:
        max_number_read_registers = (AVAILABLE_PDU_SIZE - 2) // 2
        max_numbers_write_registers = (AVAILABLE_PDU_SIZE - 10) // 2

        if not 1 <= self.number_of_read_registers <= max_number_read_registers:
            raise ValueError(
                f"Invalid number of read registers : {self.number_of_read_registers}"
            )

        if not 1 <= len(self.write_values) <= max_numbers_write_registers:
            raise ValueError(
                f"Invalid number of write registers : {self.number_of_read_registers}"
            )

        _check_address(
            self.read_starting_address,
            MAX_ADDRESS - self.number_of_read_registers + 1,
        )
        _check_address(
            self.write_starting_address, MAX_ADDRESS - len(self.write_values) + 1
        )

    def make_sdu(self) -> bytes:
        sdu = struct.pack(
            ENDIAN + f"BHHHHB{len(self.write_values)}H",
            self.function_code().value,
            _dm_to_pdu_address(self.read_starting_address),
            self.number_of_read_registers,
            _dm_to_pdu_address(self.write_starting_address),
            len(self.write_values),
            len(self.write_values) * 2,
            *self.write_values,
        )
        return sdu

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        _, byte_count = struct.unpack(ENDIAN + "BB", sdu[:2])
        expected = self.number_of_read_registers * 2
        if byte_count != expected:
            raise ProtocolReadError(
                f"Invalid byte count : {byte_count}, expected {expected}"
            )
        read_values = list(
            struct.unpack(ENDIAN + f"{self.number_of_read_registers}H", sdu[2:])
        )
        return self.Response(read_values)


@dataclass
class ReadFifoQueueSDU(ModbusRequestSDU):
    """Read FIFO queue request."""

    fifo_address: int

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        values: list[int]

    def function_code(self) -> FunctionCode:
        return FunctionCode.READ_FIFO_QUEUE

    def exceptions(self) -> dict[int, str]:
        return {
            1: "Function code not supported",
            2: "Invalid FIFO address",
            3: "Invalid FIFO count (>31)",
            4: "Couldn't read FIFO queue",
        }

    def __post_init__(self) -> None:
        _check_address(self.fifo_address)

    def make_sdu(self) -> bytes:
        return struct.pack(
            ENDIAN + "BH",
            self.function_code().value,
            _dm_to_pdu_address(self.fifo_address),
        )

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        _, byte_count = struct.unpack(ENDIAN + "BH", sdu[:3])
        fifo_count = struct.unpack(ENDIAN + "H", sdu[3:5])[0]
        register_count = byte_count // 2 - 1
        if fifo_count != register_count:
            raise ProtocolReadError(
                f"FIFO count mismatch : {fifo_count} != {register_count}"
            )
        values = list(struct.unpack(ENDIAN + f"{register_count}H", sdu[5:]))
        return self.Response(values)


@dataclass
class EncapsulatedInterfaceTransportSDU(ModbusRequestSDU):
    """Encapsulated interface transport request."""

    mei_type: int
    mei_data: bytes
    extra_exceptions: dict[int, str] | None = None

    @dataclass
    class Response(ModbusSDU):
        """Response data."""

        data: bytes

    def function_code(self) -> FunctionCode:
        return FunctionCode.ENCAPSULATED_INTERFACE_TRANSPORT

    def exceptions(self) -> dict[int, str]:
        exceptions = {1: "Function code not supported"}
        if self.extra_exceptions is not None:
            exceptions.update(self.extra_exceptions)
        return exceptions

    def make_sdu(self) -> bytes:
        return struct.pack(
            ENDIAN + f"BB{len(self.mei_data)}s",
            self.function_code().value,
            self.mei_type,
            self.mei_data,
        )

    def parse_sdu(self, sdu: bytes) -> ModbusSDU:
        self._check_for_error(sdu)
        return self.Response(sdu[2:])


class ModbusFrame(ProtocolFrame[ModbusSDU]):
    """Modbus frame containing a ModbusSDU"""

    payload: ModbusSDU

    def __str__(self) -> str:
        return f"ModbusFrame({self.payload})"


# pylint: disable=too-many-public-methods
class Modbus(Protocol[ModbusSDU]):
    """
    Modbus protocol

    Parameters
    ----------
    adapter : Adapter
        SerialPort or IP
    timeout : Timeout
    _type : str
        Only used with SerialPort adapter
        'RTU' : Modbus RTU (default)
        'ASCII' : Modbus ASCII
    """

    def __init__(
        self,
        adapter: Adapter,
        timeout: Timeout | None | EllipsisType = ...,
        _type: str = ModbusType.RTU.value,
        slave_address: int | None = None,
    ) -> None:
        super().__init__(adapter, timeout)
        self._logger.debug("Initializing Modbus protocol...")

        if isinstance(adapter, IP):
            self._adapter: IP
            self._adapter.set_default_port(MODBUS_TCP_DEFAULT_PORT)
            self._modbus_type = ModbusType.TCP
        elif isinstance(adapter, SerialPort):
            self._modbus_type = ModbusType(_type)
            if slave_address is None:
                raise ValueError("slave_address must be set")
            raise NotImplementedError("Serialport (Modbus RTU) is not supported yet")
        else:
            raise ValueError("Invalid adapter")

        self._slave_address = slave_address

        self._last_sdu: ModbusSDU | None = None
        self._transaction_id = 0

    def _default_timeout(self) -> Timeout | None:
        return Timeout(response=1, action="error")

    def _protocol_to_adapter(self, protocol_payload: ModbusSDU) -> bytes:
        if isinstance(protocol_payload, SerialLineOnlySDU):
            if self._modbus_type == ModbusType.TCP:
                raise ProtocolError("This function cannot be used with Modbus TCP")

        sdu = protocol_payload.make_sdu()

        if self._modbus_type == ModbusType.TCP:
            # Return raw data
            length = len(sdu) + 1  # unit_id is included
            output = (
                struct.pack(
                    ENDIAN + "HHHB", self._transaction_id, _PROTOCO_ID, length, _UNIT_ID
                )
                + sdu
            )
        else:
            # Add slave address and error check
            error_check = modbus_crc(sdu)
            output = (
                struct.pack(ENDIAN + "B", self._slave_address)
                + sdu
                + struct.pack(ENDIAN + "H", error_check)
            )
            if self._modbus_type == ModbusType.ASCII:
                # Add header and trailer
                output = _ASCII_HEADER + output + _ASCII_TRAILER

        self._transaction_id += 1

        self._last_sdu = protocol_payload
        return output

    def _adapter_to_protocol(
        self, adapter_frame: AdapterFrame
    ) -> ProtocolFrame[ModbusSDU]:
        pdu = adapter_frame.get_payload()

        if self._modbus_type == ModbusType.TCP:
            # transaction_id, protocol_id, length, unit_id = struct.unpack(
            #     "HHHB",
            #     pdu[:7]
            # )
            data = pdu[7:]
            # len(data) should match length variable
        else:
            if self._modbus_type == ModbusType.ASCII:
                # Remove header and trailer
                pdu = pdu[len(_ASCII_HEADER) : -len(_ASCII_TRAILER)]
            # Remove slave address and CRC and check CRC

            # slave_address = pdu[0]
            data = pdu[1:-2]
            # crc = pdu[-2:] # TODO : Check CRC

        if self._last_sdu is None:
            raise ModbusError("Cannot read without prior write")

        # It is necessary to know the previous SDU because some information cannot
        # be parsed from the response only (like the number of coils from a read_coils
        # command)
        sdu = self._last_sdu.parse_sdu(data)

        return ModbusFrame(
            stop_timestamp=adapter_frame.stop_timestamp,
            stop_condition_type=adapter_frame.stop_condition_type,
            previous_read_buffer_used=adapter_frame.previous_read_buffer_used,
            response_delay=adapter_frame.response_delay,
            payload=sdu,
        )

    # ┌────────────┐
    # │ Public API │
    # └────────────┘

    # Read Coils - 0x01
    def read_coils(self, start_address: int, number_of_coils: int) -> list[bool]:
        """
        Read a defined number of coils starting at a set address

        Parameters
        ----------
        start_address : int
        number_of_coils : int

        Returns
        -------
        coils : list
        """
        payload = ReadCoilsSDU(
            start_address=start_address, number_of_coils=number_of_coils
        )

        output = cast(ReadCoilsSDU.Response, self.query(payload))

        return output.coils

    async def aread_coils(self, start_address: int, number_of_coils: int) -> list[bool]:
        """
        Asynchronously read a defined number of coils starting at a set address

        Parameters
        ----------
        start_address : int
        number_of_coils : int

        Returns
        -------
        coils : list
        """

        payload = ReadCoilsSDU(
            start_address=start_address, number_of_coils=number_of_coils
        )

        output = cast(ReadCoilsSDU.Response, await self.aquery(payload))

        return output.coils

    # This is a wrapper for the read_coils method
    def read_single_coil(self, address: int) -> bool:
        """
        Read a single coil at a specified address.
        This is a wrapper for the read_coils method with the number of coils set to 1

        Parameters
        ----------
        address : int

        Returns
        -------
        coil : bool
        """
        coil = self.read_coils(start_address=address, number_of_coils=1)[0]
        return coil

    async def aread_single_coil(self, address: int) -> bool:
        """
        Asynchronously read a single coil at a specified address.
        This is a wrapper for the read_coils method with the number of coils set to 1

        Parameters
        ----------
        address : int

        Returns
        -------
        coil : bool
        """
        coil = (await self.aread_coils(start_address=address, number_of_coils=1))[0]
        return coil

    # Read Discrete inputs - 0x02
    def read_discrete_inputs(
        self, start_address: int, number_of_inputs: int
    ) -> list[bool]:
        """
        Read a defined number of discrete inputs at a set starting address

        Parameters
        ----------
        start_address : int
        number_of_inputs : int

        Returns
        -------
        inputs : list
        List of booleans
        """

        payload = ReadDiscreteInputs(
            start_address=start_address, number_of_inputs=number_of_inputs
        )

        output = cast(ReadDiscreteInputs.Response, self.query(payload))

        return output.inputs

    async def aread_discrete_inputs(
        self, start_address: int, number_of_inputs: int
    ) -> list[bool]:
        """
        Read a defined number of discrete inputs at a set starting address

        Parameters
        ----------
        start_address : int
        number_of_inputs : int

        Returns
        -------
        inputs : list
        List of booleans
        """

        payload = ReadDiscreteInputs(
            start_address=start_address, number_of_inputs=number_of_inputs
        )

        output = cast(ReadDiscreteInputs.Response, await self.aquery(payload))

        return output.inputs

    # Read Holding Registers - 0x03
    def read_holding_registers(
        self, start_address: int, number_of_registers: int
    ) -> list[int]:
        """
        Reads a defined number of registers starting at a set address

        Parameters
        ----------
        start_address : int
        number_of_registers : int
            1 to 125

        Returns
        -------
        registers : list
        """

        payload = ReadHoldingRegisters(
            start_address=start_address, number_of_registers=number_of_registers
        )

        output = cast(ReadHoldingRegisters.Response, self.query(payload))

        return output.registers

    async def aread_holding_registers(
        self, start_address: int, number_of_registers: int
    ) -> list[int]:
        """
        Asynchronously Reads a defined number of registers starting at a set address

        Parameters
        ----------
        start_address : int
        number_of_registers : int
            1 to 125

        Returns
        -------
        registers : list
        """

        payload = ReadHoldingRegisters(
            start_address=start_address, number_of_registers=number_of_registers
        )

        output = cast(ReadHoldingRegisters.Response, await self.aquery(payload))

        return output.registers

    def _parse_multi_register_value(
        self,
        n_registers: int,
        registers: list[int],
        value_type: str,
        *,
        byte_order: str = Endian.BIG.value,
        word_order: str = Endian.BIG.value,
        encoding: str = "utf-8",
        padding: int | None = 0,
    ) -> str | bytes | int | float:

        type_cast = TypeCast(value_type)
        _byte_order = Endian(byte_order)
        if type_cast.is_number():
            _word_order = Endian(word_order)
        else:
            _word_order = Endian.BIG
        # Create a buffer
        to_bytes_endian = Endian.BIG if _byte_order == _word_order else Endian.LITTLE
        buffer = b"".join(
            [x.to_bytes(2, byteorder=to_bytes_endian.value) for x in registers]
        )
        # Use struct_format to convert to the corresponding value directly
        # Swap the buffer accordingly
        data: bytes | int | float = struct.unpack(
            endian_symbol[_word_order] + struct_format(type_cast, n_registers * 2),
            buffer,
        )[0]

        # If data is a string, do additionnal processing
        output: bytes | int | float | str
        if type_cast == TypeCast.STRING:
            data = cast(bytes, data)
            # If null termination is enabled, remove any \0
            if padding is not None and padding in data:
                data = data[: data.index(padding)]
            # Cast
            output = data.decode(encoding)
        else:
            output = data

        return output

    def read_multi_register_value(
        self,
        address: int,
        n_registers: int,
        value_type: str,
        *,
        byte_order: str = Endian.BIG.value,
        word_order: str = Endian.BIG.value,
        encoding: str = "utf-8",
        padding: int | None = 0,
    ) -> str | bytes | int | float:
        """
        Read an integer, a float, or a string over multiple registers

        Parameters
        ----------
        address : int
            Address of the first register
        n_registers : int
            Number of registers (half the number of bytes)
        value_type : str
            Type to which the value will be cast
                'int' : signed integer
                'uint' : unsigned integer
                'float' : float or double
                'string' : string
                'array' : Bytes array
            Each type will be adapted based on the number of bytes (_bytes parameter)
        byte_order : str
            Byte order, 'big' means the high bytes will come first, 'little' means the low bytes
            will come first
            Byte order inside a register (2 bytes) is always big as per
            Modbus specification (4.2 Data Encoding)
        encoding : str
            String encoding (if used). UTF-8 by default
        padding : int | None
            String padding, None to return the raw string
        Returns
        -------
        data : any
        """
        # Read N registers
        registers = self.read_holding_registers(
            start_address=address, number_of_registers=n_registers
        )

        return self._parse_multi_register_value(
            n_registers=n_registers,
            registers=registers,
            value_type=value_type,
            byte_order=byte_order,
            word_order=word_order,
            encoding=encoding,
            padding=padding,
        )

    async def aread_multi_register_value(
        self,
        address: int,
        n_registers: int,
        value_type: str,
        *,
        byte_order: str = Endian.BIG.value,
        word_order: str = Endian.BIG.value,
        encoding: str = "utf-8",
        padding: int | None = 0,
    ) -> str | bytes | int | float:
        """
        Asynchronously read an integer, a float, or a string over multiple registers

        Parameters
        ----------
        address : int
            Address of the first register
        n_registers : int
            Number of registers (half the number of bytes)
        value_type : str
            Type to which the value will be cast
                'int' : signed integer
                'uint' : unsigned integer
                'float' : float or double
                'string' : string
                'array' : Bytes array
            Each type will be adapted based on the number of bytes (_bytes parameter)
        byte_order : str
            Byte order, 'big' means the high bytes will come first, 'little' means the low bytes
            will come first
            Byte order inside a register (2 bytes) is always big as per
            Modbus specification (4.2 Data Encoding)
        encoding : str
            String encoding (if used). UTF-8 by default
        padding : int | None
            String padding, None to return the raw string
        Returns
        -------
        data : any
        """
        # Read N registers
        registers = await self.aread_holding_registers(
            start_address=address, number_of_registers=n_registers
        )

        return self._parse_multi_register_value(
            n_registers=n_registers,
            registers=registers,
            value_type=value_type,
            byte_order=byte_order,
            word_order=word_order,
            encoding=encoding,
            padding=padding,
        )

    def _make_multi_register_value(
        self,
        n_registers: int,
        value_type: str,
        value: str | bytes | int | float,
        *,
        byte_order: str = Endian.BIG.value,
        word_order: str = Endian.BIG.value,
        encoding: str = "utf-8",
        padding: int = 0,
    ) -> list[int]:
        _type = TypeCast(value_type)
        n_bytes = n_registers * 2
        if _type.is_number():
            _word_order = Endian(word_order)
        else:
            _word_order = Endian.BIG
        _byte_order = Endian(byte_order)

        array = b""
        if _type in [TypeCast.INT, TypeCast.UINT, TypeCast.FLOAT]:
            # Make one big array using word_order endian
            array = struct.pack(
                endian_symbol[_word_order] + struct_format(_type, n_bytes), value
            )

        elif _type == TypeCast.ARRAY:
            if isinstance(value, bytes):
                if len(value) > n_registers * 2:
                    raise ValueError(
                        f"Cannot store {len(value)} bytes array in {n_registers} registers"
                    )
            else:
                raise ValueError(f"Invalid value type : {type(value)}")

            array = value
        elif _type == TypeCast.STRING:
            if isinstance(value, str):
                array = value.encode(encoding)
            else:
                raise ValueError(f"Invalid value type : {type(value)}")

            if len(array) < n_bytes:
                # Padding
                array = array + padding.to_bytes(1, byteorder="big") * (
                    n_bytes - len(value)
                )

        if len(array) != n_registers * 2:
            raise ValueError(
                f"Cannot store a {len(array)} bytes array in {n_registers} registers"
            )

        unpack_endian = Endian.BIG if _byte_order == _word_order else Endian.LITTLE
        registers = [
            struct.unpack(endian_symbol[unpack_endian] + "H", array[2 * i : 2 * i + 2])[
                0
            ]
            for i in range(len(array) // 2)
        ]

        return registers

    def write_multi_register_value(
        self,
        address: int,
        n_registers: int,
        value_type: str,
        value: str | bytes | int | float,
        *,
        byte_order: str = Endian.BIG.value,
        word_order: str = Endian.BIG.value,
        encoding: str = "utf-8",
        padding: int = 0,
    ) -> None:
        """
        Write an integer, a float, or a string over multiple registers

        Parameters
        ----------
        address : int
            Address of the first register
        n_registers : int
            Number of registers (half the number of bytes)
        value_type : str | bytes | int
            Type to which the value will be cast
                'int' : signed integer
                'uint' : unsigned integer
                'float' : float or double
                'string' : string
                'array' : Bytes array
        value : any
            The value to write, can be any of the following : str, int, float, str, bytearray
        byte_order : str
            Byte order, 'big' means the high bytes will come first, 'little' means the low
            bytes will come first
            Byte order inside a register (2 bytes) is always big as per
            Modbus specification (4.2 Data Encoding)
        encoding : str
            String encoding (if used)
        padding : int
            Padding in case the value (str / bytes) is not long enough
        Returns
        -------
        data : any
        """

        registers = self._make_multi_register_value(
            n_registers=n_registers,
            value_type=value_type,
            value=value,
            byte_order=byte_order,
            word_order=word_order,
            encoding=encoding,
            padding=padding,
        )

        self.write_multiple_registers(start_address=address, values=registers)

    async def awrite_multi_register_value(
        self,
        address: int,
        n_registers: int,
        value_type: str,
        value: str | bytes | int | float,
        *,
        byte_order: str = Endian.BIG.value,
        word_order: str = Endian.BIG.value,
        encoding: str = "utf-8",
        padding: int = 0,
    ) -> None:
        """
        Asynchronously write an integer, a float, or a string over multiple registers

        Parameters
        ----------
        address : int
            Address of the first register
        n_registers : int
            Number of registers (half the number of bytes)
        value_type : str | bytes | int
            Type to which the value will be cast
                'int' : signed integer
                'uint' : unsigned integer
                'float' : float or double
                'string' : string
                'array' : Bytes array
        value : any
            The value to write, can be any of the following : str, int, float, str, bytearray
        byte_order : str
            Byte order, 'big' means the high bytes will come first, 'little' means the low
            bytes will come first
            Byte order inside a register (2 bytes) is always big as per
            Modbus specification (4.2 Data Encoding)
        encoding : str
            String encoding (if used)
        padding : int
            Padding in case the value (str / bytes) is not long enough
        Returns
        -------
        data : any
        """

        registers = self._make_multi_register_value(
            n_registers=n_registers,
            value_type=value_type,
            value=value,
            byte_order=byte_order,
            word_order=word_order,
            encoding=encoding,
            padding=padding,
        )

        await self.awrite_multiple_registers(start_address=address, values=registers)

    # Read Input Registers - 0x04
    def read_input_registers(
        self, start_address: int, number_of_registers: int
    ) -> list[int]:
        """
        Reads a defined number of input registers starting at a set address

        Parameters
        ----------
        start_address : int
        number_of_registers : int
            1 to 125

        Returns
        -------
        registers : list
            List of integers
        """
        payload = ReadInputRegistersSDU(
            start_address=start_address,
            number_of_registers=number_of_registers,
        )

        output = cast(ReadInputRegistersSDU.Response, self.query(payload))

        return output.registers

    async def aread_input_registers(
        self, start_address: int, number_of_registers: int
    ) -> list[int]:
        """
        Asynchronously read a defined number of input registers starting at a set address

        Parameters
        ----------
        start_address : int
        number_of_registers : int
            1 to 125

        Returns
        -------
        registers : list
            List of integers
        """
        payload = ReadInputRegistersSDU(
            start_address=start_address,
            number_of_registers=number_of_registers,
        )

        output = cast(ReadInputRegistersSDU.Response, await self.aquery(payload))

        return output.registers

    # Write Single coil - 0x05
    def write_single_coil(self, address: int, status: bool) -> None:
        """
        Write a single output to either ON or OFF

        Parameters
        ----------
        address : int
        status : bool
        """
        payload = WriteSingleCoilSDU(address=address, status=status)

        self.query(payload)

    async def awrite_single_coil(self, address: int, status: bool) -> None:
        """
        Asynchronously write a single output to either ON or OFF

        Parameters
        ----------
        address : int
        status : bool
        """
        payload = WriteSingleCoilSDU(address=address, status=status)

        await self.aquery(payload)

    # Write Single Register - 0x06
    def write_single_register(self, address: int, value: int) -> None:
        """
        Write a single register

        Parameters
        ----------
        address : int
        value : int
            value between 0x0000 and 0xFFFF
        """
        payload = WriteSingleRegisterSDU(
            address=address,
            value=value,
        )

        self.query(payload)

    async def awrite_single_register(self, address: int, value: int) -> None:
        """
        Asynchronously write a single register

        Parameters
        ----------
        address : int
        value : int
            value between 0x0000 and 0xFFFF
        """
        payload = WriteSingleRegisterSDU(
            address=address,
            value=value,
        )

        await self.aquery(payload)

    def read_single_register(self, address: int) -> int:
        """
        Read a single register

        Parameters
        ----------
        address : int

        Returns
        -------
        value : int
        """
        return self.read_holding_registers(address, 1)[0]

    # Read Exception Status - 0x07
    def read_exception_status(self) -> int:
        """
        Read exeption status

        Returns
        -------
        exceptions : int
        """
        payload = ReadExceptionStatusSDU()

        output = cast(ReadExceptionStatusSDU.Response, self.query(payload))

        return output.status

    async def aread_exception_status(self) -> int:
        """
        Asynchronously read exeption status

        Returns
        -------
        exceptions : int
        """
        payload = ReadExceptionStatusSDU()

        output = cast(ReadExceptionStatusSDU.Response, await self.aquery(payload))

        return output.status

    # Diagnostics - 0x08
    def diagnostics_return_query_data(self, data: int = 0x1234) -> bool:
        """
        Run "Return Query Data" diagnostic

        A query is sent and should be return identical

        Parameters
        ----------
        data : int
            data to send (16 bits integer)

        Returns
        -------
        success : bool
        """
        subfunction_data = struct.pack(ENDIAN + "H", data)
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_QUERY_DATA,
            subfunction_data=subfunction_data,
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, self.query(payload))

        return output.data == subfunction_data

    async def adiagnostics_return_query_data(self, data: int = 0x1234) -> bool:
        """
        Asynchronously run "Return Query Data" diagnostic

        A query is sent and should be return identical

        Parameters
        ----------
        data : int
            data to send (16 bits integer)

        Returns
        -------
        success : bool
        """
        subfunction_data = struct.pack(ENDIAN + "H", data)
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_QUERY_DATA,
            subfunction_data=subfunction_data,
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, await self.aquery(payload))

        return output.data == subfunction_data

    # TODO : Check how this function interracts with Listen Only mode
    def diagnostics_restart_communications_option(
        self, clear_communications_event_log: bool = False
    ) -> None:
        """
        Initialize and restart serial line port. Brings the device out of Listen Only Mode

        Parameters
        ----------
        clear_communications_event_log : bool
            False by default
        """
        subfunction_data = struct.pack(
            ENDIAN + "H", 0xFF00 if clear_communications_event_log else 0x0000
        )
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RESTART_COMMUNICATIONS_OPTION,
            subfunction_data=subfunction_data,
            return_subfunction_bytes=2,
        )

        self.query(payload)

    async def adiagnostics_restart_communications_option(
        self, clear_communications_event_log: bool = False
    ) -> None:
        """
        Asynchronously initialize and restart serial line port.
        Brings the device out of Listen Only Mode

        Parameters
        ----------
        clear_communications_event_log : bool
            False by default
        """
        subfunction_data = struct.pack(
            ENDIAN + "H", 0xFF00 if clear_communications_event_log else 0x0000
        )
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RESTART_COMMUNICATIONS_OPTION,
            subfunction_data=subfunction_data,
            return_subfunction_bytes=2,
        )

        await self.aquery(payload)

    def diagnostics_return_diagnostic_register(self) -> int:
        """
        Return 16 bit diagnostic register

        Returns
        -------
        register : int
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_DIAGNOSTIC_REGISTER,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, self.query(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    async def adiagnostics_return_diagnostic_register(self) -> int:
        """
        Asynchronously return 16 bit diagnostic register

        Returns
        -------
        register : int
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_DIAGNOSTIC_REGISTER,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, await self.aquery(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    def diagnostics_change_ascii_input_delimiter(self, char: bytes | str) -> None:
        """
        Change the ASCII input delimiter to specified value

        Parameters
        ----------
        char : bytes or str
            Single character
        """
        if len(char) != 1:
            raise ValueError(f"Invalid char length : {len(char)}")
        if isinstance(char, str):
            char = char.encode("ASCII")

        subfunction_data = struct.pack("cB", char, 0)
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.CHANGE_ASCII_INPUT_DELIMITER,
            subfunction_data=subfunction_data,
            return_subfunction_bytes=2,
        )

        self.query(payload)

    async def adiagnostics_change_ascii_input_delimiter(
        self, char: bytes | str
    ) -> None:
        """
        Asynchronously change the ASCII input delimiter to specified value

        Parameters
        ----------
        char : bytes or str
            Single character
        """
        if len(char) != 1:
            raise ValueError(f"Invalid char length : {len(char)}")
        if isinstance(char, str):
            char = char.encode("ASCII")

        subfunction_data = struct.pack("cB", char, 0)
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.CHANGE_ASCII_INPUT_DELIMITER,
            subfunction_data=subfunction_data,
            return_subfunction_bytes=2,
        )

        await self.aquery(payload)

    def diagnostics_force_listen_only_mode(self) -> None:
        """
        Forces the addressed remote device to its Listen Only Mode for MODBUS communications.
        This isolates it from the other devices on the network, allowing them to continue
        communicating without interruption from the addressed remote device. No response is
        returned.
        When the remote device enters its Listen Only Mode, all active communication controls are
        turned off. The Ready watchdog timer is allowed to expire, locking the controls off.
        While the device is in this mode, any MODBUS messages addressed to it or broadcast
        are monitored, but no actions will be taken and no responses will be sent.
        The only function that will be processed after the mode is entered will be the Restart
        Communications Option function (function code 8, sub-function 1).
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.FORCE_LISTEN_ONLY_MODE,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=0,
            check_response=False,
        )

        self.query(payload)

    async def adiagnostics_force_listen_only_mode(self) -> None:
        """
        Asynchronously force the addressed remote device to its Listen Only Mode
        for MODBUS communications.
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.FORCE_LISTEN_ONLY_MODE,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=0,
            check_response=False,
        )

        await self.aquery(payload)

    def diagnostics_clear_counters_and_diagnostic_register(self) -> None:
        """
        Clear all counters and the diagnostic register
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.CLEAR_COUNTERS_AND_DIAGNOSTIC_REGISTER,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=0,
        )

        self.query(payload)

    async def adiagnostics_clear_counters_and_diagnostic_register(self) -> None:
        """
        Asynchronously clear all counters and the diagnostic register
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.CLEAR_COUNTERS_AND_DIAGNOSTIC_REGISTER,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=0,
        )

        await self.aquery(payload)

    def diagnostics_return_bus_message_count(self) -> int:
        """
        Return the number of messages that the remote device has detection on the communications
        system since its last restat, clear counters operation, or power-up

        Returns
        -------
        count : int
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_BUS_MESSAGE_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, self.query(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    async def adiagnostics_return_bus_message_count(self) -> int:
        """
        Asynchronously return the number of messages that the remote device has detection on the
        communications system since its last restat, clear counters operation, or power-up
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_BUS_MESSAGE_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, await self.aquery(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    def diagnostics_return_bus_communication_error_count(self) -> int:
        """
        Return the number of messages that the remote device has detection on the communications
        system since its last restart, clear counters operation, or power-up

        Returns
        -------
        count : int
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_BUS_COMMUNICATION_ERROR_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, self.query(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    async def adiagnostics_return_bus_communication_error_count(self) -> int:
        """
        Asynchronously return the number of messages that the remote device has detection on the
        communications system since its last restart, clear counters operation, or power-up
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_BUS_COMMUNICATION_ERROR_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, await self.aquery(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    def diagnostics_return_bus_exception_error_count(self) -> int:
        """
        Return the number of Modbus exceptions responses returned by the remote device since
        its last restart, clear counters operation, or power-up

        Returns
        -------
        count : int
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_BUS_EXCEPTION_ERROR_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, self.query(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    async def adiagnostics_return_bus_exception_error_count(self) -> int:
        """
        Asynchronously return the number of Modbus exceptions responses returned by the remote
        device since its last restart, clear counters operation, or power-up
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_BUS_EXCEPTION_ERROR_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, await self.aquery(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    def diagnostics_return_server_no_response_count(self) -> int:
        """
        Return the number of messages addressed to the remote device for which it has returned
        no response since its last restart, clear counters operation, or power-up

        Returns
        -------
        count : int
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_SERVER_NO_RESPONSE_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, self.query(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    async def adiagnostics_return_server_no_response_count(self) -> int:
        """
        Asynchronously return the number of messages addressed to the remote device for which it
        has returned no response since its last restart, clear counters operation, or power-up
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_SERVER_NO_RESPONSE_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, await self.aquery(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    def diagnostics_return_server_nak_count(self) -> int:
        """
        Return the number of messages addressed to the remote device for which it returned
        a negative acnowledge (NAK) exception response since its last restart, clear counters
        operation, or power-up

        Returns
        -------
        count : int
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_SERVER_NAK_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, self.query(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    async def adiagnostics_return_server_nak_count(self) -> int:
        """
        Asynchronously return the number of messages addressed to the remote device for which it
        returned a negative acnowledge (NAK) exception response since its last restart,
        clear counters operation, or power-up
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_SERVER_NAK_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, await self.aquery(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    def diagnostics_return_server_busy_count(self) -> int:
        """
        Return the number of messages addressed to the remote device for which it returned a
        server device busy exception response since its last restart, clear counters operation,
        or power-up

        Returns
        -------
        count : int
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_SERVER_BUSY_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, self.query(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    async def adiagnostics_return_server_busy_count(self) -> int:
        """
        Asynchronously return the number of messages addressed to the remote device for which it
        returned a server device busy exception response since its last restart, clear counters
        operation, or power-up
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_SERVER_BUSY_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, await self.aquery(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    def diagnostics_return_bus_character_overrun_count(self) -> int:
        """
        Return the number of messages addressed to the remote device that it could not handle
        due to a character overrun condition since its last restart, clear counters operation,
        or power-up

        Returns
        -------
        count : int
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_BUS_CHARACTER_OVERRUN_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, self.query(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    async def adiagnostics_return_bus_character_overrun_count(self) -> int:
        """
        Asynchronously return the number of messages addressed to the remote device that it could
        not handle due to a character overrun condition since its last restart, clear counters
        operation, or power-up
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.RETURN_BUS_CHARACTER_OVERRUN_COUNT,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=2,
        )

        output = cast(DiagnosticsSDU.Response, await self.aquery(payload))

        return int(struct.unpack(ENDIAN + "H", output.data)[0])

    def diagnostics_clear_overrun_counter_and_flag(self) -> None:
        """
        Clear the overrun error counter and reset the error flag
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.CLEAR_OVERRUN_COUNTER_AND_FLAG,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=0,
        )

        self.query(payload)

    async def adiagnostics_clear_overrun_counter_and_flag(self) -> None:
        """
        Asynchronously clear the overrun error counter and reset the error flag
        """
        payload = DiagnosticsSDU(
            code=DiagnosticsCode.CLEAR_OVERRUN_COUNTER_AND_FLAG,
            subfunction_data=b"\x00\x00",
            return_subfunction_bytes=0,
        )

        await self.aquery(payload)

    # Get Comm Event Counter - 0x0B
    def get_comm_event_counter(self) -> tuple[int, int]:
        """
        Retrieve status word and event count from the remote device's communication event counter

        Returns
        -------
        status : int
        event_count : int
        """
        payload = GetCommEventCounterSDU()

        output = cast(GetCommEventCounterSDU.Response, self.query(payload))

        return output.status, output.event_count

    async def aget_comm_event_counter(self) -> tuple[int, int]:
        """
        Asynchronously retrieve status word and event count from the remote device's
        communication event counter
        """
        payload = GetCommEventCounterSDU()

        output = cast(GetCommEventCounterSDU.Response, await self.aquery(payload))

        return output.status, output.event_count

    # Get Comm Event Log - 0x0C
    def get_comm_event_log(self) -> tuple[int, int, int, bytes]:
        """
        Retrieve status word, event count, message count and a field of event bytes from
        the remote device

        Status word and event count are identical to those returned by get_comm_event_counter()

        Returns
        -------
        status : int
        event_count : int
        message_count : int
            Number of messages processed since its last restart, clear counters operation,
            or power-up
            Identical to diagnostics_return_bus_message_count()
        events : bytes
            0-64 bytes, each corresponding to the status of one Modbus send or receive
            operation for the remote device. Byte 0 is the most recent event
        """
        payload = GetCommEventLogSDU()

        output = cast(GetCommEventLogSDU.Response, self.query(payload))

        return output.status, output.event_count, output.message_count, output.events

    async def aget_comm_event_log(self) -> tuple[int, int, int, bytes]:
        """
        Asynchronously retrieve status word, event count, message count and a field of event bytes
        from the remote device
        """
        payload = GetCommEventLogSDU()

        output = cast(GetCommEventLogSDU.Response, await self.aquery(payload))

        return output.status, output.event_count, output.message_count, output.events

    # Write Multiple Coils - 0x0F
    def write_multiple_coils(self, start_address: int, values: list[bool]) -> None:
        """
        Write multiple coil values

        Parameters
        ----------
        start_address : int
        value : list
            Bool values
        """
        payload = WriteMultipleCoilsSDU(
            start_address=start_address,
            values=values,
        )

        self.query(payload)

    async def awrite_multiple_coils(
        self, start_address: int, values: list[bool]
    ) -> None:
        """
        Asynchronously write multiple coil values
        """
        payload = WriteMultipleCoilsSDU(
            start_address=start_address,
            values=values,
        )

        await self.aquery(payload)

    # Write Multiple Registers - 0x10
    def write_multiple_registers(self, start_address: int, values: list[int]) -> None:
        """
        Write multiple registers

        Parameters
        ----------
        start_address : int
        values : list
            List of integers

        """
        payload = WriteMultipleRegistersSDU(
            start_address=start_address,
            values=values,
        )

        self.query(payload)

    async def awrite_multiple_registers(
        self, start_address: int, values: list[int]
    ) -> None:
        """
        Asynchronously write multiple registers
        """
        payload = WriteMultipleRegistersSDU(
            start_address=start_address,
            values=values,
        )

        await self.aquery(payload)

    # Report Server ID - 0x11
    def report_server_id(
        self, server_id_length: int, additional_data_length: int
    ) -> tuple[bytes, bool, bytes]:
        """
        Read description of the type, current status and other information specific to
        a remote device

        Parameters
        ----------
        server_id_length : int
            Length of server id field (bytes)
        additional_data_length : int
            Length of additional data (bytes), 0 if none
        Returns
        -------
        server_id : bytes
        run_indicator_status : bool
        additional_data : bytes
        """
        payload = ReportServerIdSDU(
            server_id_length=server_id_length,
            additional_data_length=additional_data_length,
        )

        output = cast(ReportServerIdSDU.Response, self.query(payload))

        return output.server_id, output.run_indicator_status, output.additional_data

    async def areport_server_id(
        self, server_id_length: int, additional_data_length: int
    ) -> tuple[bytes, bool, bytes]:
        """
        Asynchronously read description of the type, current status and other information specific
        to a remote device
        """
        payload = ReportServerIdSDU(
            server_id_length=server_id_length,
            additional_data_length=additional_data_length,
        )

        output = cast(ReportServerIdSDU.Response, await self.aquery(payload))

        return output.server_id, output.run_indicator_status, output.additional_data

    # Read File Record - 0x14
    def read_file_record(self, records: list[tuple[int, int, int]]) -> list[bytes]:
        """
        Perform a single or multiple file record read

        Total response length cannot exceed 253 bytes, meaning the number of records
        and their length is limited.


        Query equation : 2 + 7*N <= 253
        Response equation : 2 + N*2 + sum(Li*2) <= 253

        Parameters
        ----------
        records : list
            List of tuples : (file_number, record_number, record_length)

        Returns
        -------
        records_data : list
            List of bytes
        """
        payload = ReadFileRecordSDU(records=records)

        output = cast(ReadFileRecordSDU.Response, self.query(payload))

        return output.records_data

    async def aread_file_record(
        self, records: list[tuple[int, int, int]]
    ) -> list[bytes]:
        """
        Asynchronously perform a single or multiple file record read
        """
        payload = ReadFileRecordSDU(records=records)

        output = cast(ReadFileRecordSDU.Response, await self.aquery(payload))

        return output.records_data

    # Write File Record - 0x15
    def write_file_record(self, records: list[tuple[int, int, bytes]]) -> None:
        """
        Perform a single or multiple file record write

        Total query and response length cannot exceed 253 bytes, meaning the number of
        records and their length is limited.

        Query equation : 2 + 7*N + sum(Li*2) <= 253
        Response equation : identical

        File number can be between 0x0001 and 0xFFFF but lots of legacy equipment will
        not support file number above 0x000A (10)

        Parameters
        ----------
        records : list
            List of tuples : (file_number, record_number, data)
        """
        payload = WriteFileRecordSDU(records=records)

        self.query(payload)

    async def awrite_file_record(self, records: list[tuple[int, int, bytes]]) -> None:
        """
        Asynchronously perform a single or multiple file record write
        """
        payload = WriteFileRecordSDU(records=records)

        await self.aquery(payload)

    # Mask Write Register - 0x16
    async def amask_write_register(
        self, address: int, and_mask: int, or_mask: int
    ) -> None:
        """
        This function is used to modify the contents of a holding register using a
        combination of AND and OR masks applied to the current contents of the register.

        The algorithm is :

        New value = (old value & and_mask) | (or_mask & (~and_mask))

        Parameters
        ----------
        address : int
            0x0000 to 0xFFFF
        and_mask : int
            0x0000 to 0xFFFF
        or_mask : int
            0x0000 to 0xFFFF
        """
        payload = MaskWriteRegisterSDU(
            address=address,
            and_mask=and_mask,
            or_mask=or_mask,
        )

        await self.aquery(payload)

    def mask_write_register(self, address: int, and_mask: int, or_mask: int) -> None:
        """
        This function is used to modify the contents of a holding register using a
        combination of AND and OR masks applied to the current contents of the register.
        """
        payload = MaskWriteRegisterSDU(
            address=address,
            and_mask=and_mask,
            or_mask=or_mask,
        )

        self.query(payload)

    # Read/Write Multiple Registers - 0x17
    def read_write_multiple_registers(
        self,
        read_starting_address: int,
        number_of_read_registers: int,
        write_starting_address: int,
        write_values: list[int],
    ) -> list[int]:
        """
        Do a write, then a read operation, each on a specific set of registers.

        Parameters
        ----------
        read_starting_address : int
        number_of_read_registers : int
        write_starting_address : int
        write_values : list
            List of registers values

        Returns
        -------
        read_values : list
        """
        payload = ReadWriteMultipleRegistersSDU(
            read_starting_address=read_starting_address,
            number_of_read_registers=number_of_read_registers,
            write_starting_address=write_starting_address,
            write_values=write_values,
        )

        output = cast(ReadWriteMultipleRegistersSDU.Response, self.query(payload))

        return output.read_values

    async def aread_write_multiple_registers(
        self,
        read_starting_address: int,
        number_of_read_registers: int,
        write_starting_address: int,
        write_values: list[int],
    ) -> list[int]:
        """
        Asynchronously do a write, then a read operation, each on a specific set of registers.
        """
        payload = ReadWriteMultipleRegistersSDU(
            read_starting_address=read_starting_address,
            number_of_read_registers=number_of_read_registers,
            write_starting_address=write_starting_address,
            write_values=write_values,
        )

        output = cast(
            ReadWriteMultipleRegistersSDU.Response, await self.aquery(payload)
        )

        return output.read_values

    # Read FIFO Queue - 0x18
    def read_fifo_queue(self, fifo_address: int) -> list[int]:
        """
        Read the contents of a First-In-First-Out (FIFO) queue of registers

        Parameters
        ----------
        fifo_address : int

        Returns
        -------
        registers : list
        """
        payload = ReadFifoQueueSDU(fifo_address=fifo_address)

        output = cast(ReadFifoQueueSDU.Response, self.query(payload))

        return output.values

    async def aread_fifo_queue(self, fifo_address: int) -> list[int]:
        """
        Asynchronously read the contents of a First-In-First-Out (FIFO) queue of registers
        """
        payload = ReadFifoQueueSDU(fifo_address=fifo_address)

        output = cast(ReadFifoQueueSDU.Response, await self.aquery(payload))

        return output.values

    # Encapsulate Interface Transport - 0x2B
    def encapsulated_interface_transport(
        self,
        mei_type: int,
        mei_data: bytes,
        extra_exceptions: dict[int, str] | None = None,
    ) -> bytes:
        """
        The MODBUS Encapsulated Interface (MEI) Transport is a mechanism for tunneling
        service requests and method invocations

        Parameters
        ----------
        mei_type : int
        mei_data : bytes

        Returns
        -------
        returned_mei_data : bytes
        """
        payload = EncapsulatedInterfaceTransportSDU(
            mei_type=mei_type,
            mei_data=mei_data,
            extra_exceptions=extra_exceptions,
        )

        output = cast(EncapsulatedInterfaceTransportSDU.Response, self.query(payload))

        return output.data

    async def aencapsulated_interface_transport(
        self,
        mei_type: int,
        mei_data: bytes,
        extra_exceptions: dict[int, str] | None = None,
    ) -> bytes:
        """
        Asynchronously use the MODBUS Encapsulated Interface (MEI) Transport
        """
        payload = EncapsulatedInterfaceTransportSDU(
            mei_type=mei_type,
            mei_data=mei_data,
            extra_exceptions=extra_exceptions,
        )

        output = cast(
            EncapsulatedInterfaceTransportSDU.Response, await self.aquery(payload)
        )

        return output.data



DEFAULT_REMOTE_PORT = 59677


# Backend protocol format
# (action, arguments)
# If the command succeeds, it is returned as :
# (action, other arguments)
# If the command fails, it is returned as :
# (error, error_description)


# The backend links the client with an adapter when SELECT_ADAPTER is
# sent along with an adapter descriptor

EXTRA_BUFFER_RESPONSE_TIME = 1

# Delay to let the adapter connect
DEFAULT_ADAPTER_OPEN_TIMEOUT = 0.5


# def add_backend_address_port_arguments(
#     parser: argparse.ArgumentParser, client_side: bool
# ) -> None:
#     """
#     Add -a/--address and -p/--port to a given ArgumentParser. The description of -a/--address
#     is different based on the value of client_side
#     """
#     if client_side:
#         address_help = "Address of the backend"
#     else:
#         address_help = (
#             "Backend listening address, set it to the host address used by clients"
#         )

#     parser.add_argument(
#         "-a", "--address", type=str, default=LOCALHOST, help=address_help
#     )
#     parser.add_argument("-p", "--port", type=int, default=BACKEND_PORT)


# class Action(Enum):
#     """
#     Backend actions enum
#     """

#     # All adapters
#     SELECT_ADAPTER = "select"
#     OPEN = "open"  # (descriptor,stop_condition) -> ()
#     CLOSE = "close"  # (descriptor,force) -> ()
#     # FORCE_CLOSE = "force_close"  # (descriptor,) -> ()
#     WRITE = "write"  # (descriptor,data) -> ()
#     READ = "read"  # (descriptor,full_output,temp_timeout,temp_stop_condition) -> (data,metrics)
#     SET_STOP_CONDITIONS = "set_stop_condition"  # (descriptor,stop_condition)
#     FLUSHREAD = "flushread"
#     START_READ = "start_read"  # Start a read (descriptor,response_time)
#     RESPONSE_TIMEOUT = "response_timeout"

#     # Signal
#     ADAPTER_SIGNAL = "adapter_signal"

#     # Other
#     SET_ROLE_ADAPTER = (
#         "set_role_adapter"  # Define the client as an adapter (exchange of data)
#     )
#     SET_ROLE_MONITORING = "set_role_monitoring"  # The client queries for backend info
#     SET_ROLE_LOGGER = "set_role_logger"  # The client receives logs
#     SET_LOG_LEVEL = "set_log_level"
#     PING = "ping"
#     STOP = "stop"

#     # Backend debugger
#     ENUMERATE_ADAPTER_CONNECTIONS = "enumerate_adapter_connections"
#     ENUMERATE_MONITORING_CONNECTIONS = "enumerate_monitoring_connections"
#     BACKEND_STATS = "backend_stats"

#     # Errors
#     ERROR_GENERIC = "error_generic"
#     ERROR_UNKNOWN_ACTION = "error_unknown_action"
#     ERROR_INVALID_REQUEST = "error_invalid_request"
#     ERROR_ADAPTER_NOT_OPENED = "error_adapter_not_opened"
#     ERROR_INVALID_ROLE = "error_invalid_role"
#     ERROR_ADAPTER_DISCONNECTED = "error_adapter_disconnected"
#     ERROR_BACKEND_DISCONNECTED = "error_backend_disconnected"
#     ERROR_FAILED_TO_OPEN = "error_failed_to_open"


# def is_action_error(action: Action) -> bool:
#     """
#     Return True if the action describes an error
#     """
#     return action.value.startswith("error_")


# class BackendException(Exception):
#     """
#     Generic backend error
#     """


# def frontend_send(conn: Connection, action: Action, *args: Any) -> bool:
#     """
#     Send an action to the frontend
#     """
#     try:
#         conn.send((action.value, *args))
#     except (BrokenPipeError, OSError):
#         return False
#     return True


# def backend_request(
#     conn: Connection,
#     action: Action,
#     *args: Any,
#     timeout: float = EXTRA_BUFFER_RESPONSE_TIME,
# ) -> BackendResponse:
#     """
#     Send a request to the backend an expect a response with the time specified by timeout
#     """
#     try:
#         conn.send((action.value, *args))
#     except (BrokenPipeError, OSError) as err:
#         raise BackendCommunicationError("Failed to communicate with backend") from err

#     ready = wait([conn], timeout=timeout)
#     if conn not in ready:
#         raise BackendCommunicationError("Failed to receive backend response in time")

#     try:
#         raw_response: object = conn.recv()
#     except (EOFError, ConnectionResetError) as err:
#         raise BackendCommunicationError(
#             f"Failed to receive backend response to {action.value}"
#         ) from err

#     # Check if the response is correctly formatted
#     if not (isinstance(raw_response, tuple) and isinstance(raw_response[0], str)):
#         raise BackendCommunicationError(
#             f"Invalid response received from backend : {raw_response}"
#         )

#     response_action: Action = Action(raw_response[0])
#     arguments: tuple[Any, ...] = raw_response[1:]

#     if is_action_error(response_action):
#         if len(arguments) > 0:
#             if isinstance(arguments[0], str):
#                 error_message: str = arguments[0]
#             else:
#                 error_message = "failed to read error message"
#         else:
#             error_message = "Missing error message"
#         raise BackendError(f"{response_action} : {error_message}")
#     return arguments


# backend_send = frontend_send


# def raise_if_error(response: BackendResponse) -> None:
#     """
#     Raise error if the action is an error, ignore otherwise
#     """
#     action = Action(response[0])
#     if is_action_error(action):
#         if len(response) > 1:
#             description = response[1]
#         else:
#             description = f"{action}"
#         raise BackendException(f"{action.name}/{description}")


# class AdapterBackendStatus(Enum):
#     """
#     Adapter backend status enum
#     """

#     DISCONNECTED = 0
#     CONNECTED = 1


# class ClientStatus(Enum):
#     """
#     Client status enum
#     """

#     DISCONNECTED = 0
#     CONNECTED = 1



# File : stop_condition.py
# Author : Sébastien Deriaz
# License : GPL

"""
Stop-condition module

This is the frontend of the stop-conditions, the part that is imported by the user
"""

# from abc import abstractmethod
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum


@dataclass
class Fragment:
    """
    Fragment class, holds a piece of data (bytes) and the time at which it was received
    """

    data: bytes
    timestamp: float

    def __str__(self) -> str:
        return f"{self.data!r}@{self.timestamp}"

    def __repr__(self) -> str:
        return f"Fragment({self.data!r}@{self.timestamp})"

    def __getitem__(self, key: slice) -> "Fragment":
        # if self.data is None:
        #     raise IndexError('Cannot index invalid fragment')
        return Fragment(self.data[key], self.timestamp)


class StopConditionType(Enum):
    """
    Stop-condition type
    """
    TERMINATION = "termination"
    LENGTH = "length"
    CONTINUATION = "continuation"
    TOTAL = "total"
    FRAGMENT = "fragment"
    TIMEOUT = "timeout"


class StopCondition:
    """
    Stop-condition base class, cannot be used on its own
    """

    # @abstractmethod
    # def type(self) -> StopConditionType:
    #     pass

    @abstractmethod
    def initiate_read(self, timestamp: float) -> None:
        """
        Prepare the stop-condition for the next read
        """

    @abstractmethod
    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        """
        Evaluate incoming fragment and return read information for the next fragment
        """

    @abstractmethod
    def type(self) -> StopConditionType:
        """
        Helper function to determine the which type of stop-condition generated a stop
        """

    @abstractmethod
    def flush_read(self) -> None:
        """
        Reset read operation
        """


class Termination(StopCondition):
    """
    Termination stop-condition, used to stop when a specified sequence is received

    Parameters
    ----------
    sequence : bytes | str
    """

    def __init__(self, sequence: bytes | str) -> None:
        super().__init__()
        if isinstance(sequence, str):
            self._sequence = sequence.encode("utf-8")
        else:
            self._sequence = sequence

        if self._sequence == b"":
            raise ValueError(
                "Empty termination isn't allowed. If you wish "
                "to stop on any received data, use Datagram stop-conditions instead"
            )
        self._sequence_found_length = 0

    def __str__(self) -> str:
        return f"Termination({repr(self._sequence)})"

    def __repr__(self) -> str:
        return self.__str__()

    def initiate_read(self, timestamp: float) -> None:
        self._sequence_found_length = 0

    def flush_read(self) -> None:
        self._sequence_found_length = 0

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        if raw_fragment.data is None:
            raise RuntimeError("Trying to evaluate an invalid fragment")

        position, length = termination_in_data(
            self._sequence[self._sequence_found_length :], raw_fragment.data
        )
        stop = False
        deferred = Fragment(b"", raw_fragment.timestamp)

        if position is None:
            # Nothing was found, keep everything
            kept = raw_fragment
        else:
            self._sequence_found_length += length

            if self._sequence_found_length == len(self._sequence):
                # The sequence was found entirely
                deferred = raw_fragment[position + length :]
                self._sequence_found_length = 0
                stop = True
            elif position + length == len(raw_fragment.data):
                # Part of the sequence was found at the end
                # Return what's before the sequence
                deferred = Fragment(b"", raw_fragment.timestamp)

            kept = raw_fragment[: position + length]

        return stop, kept, deferred, None

    def type(self) -> StopConditionType:
        return StopConditionType.TERMINATION


class Length(StopCondition):
    """
    Length stop-condition, used to stop when the specified number of bytes (or more) have been read

    Parameters
    ----------
    n : int
        Number of bytes
    """

    # TYPE = StopConditionType.LENGTH
    def __init__(self, n: int) -> None:
        super().__init__()
        self._n = n
        self._counter = 0

    def __str__(self) -> str:
        return f"Length({self._n})"

    def __repr__(self) -> str:
        return self.__str__()

    def initiate_read(self, timestamp: float) -> None:
        # Length
        self._counter = 0

    def type(self) -> StopConditionType:
        return StopConditionType.LENGTH

    def flush_read(self) -> None:
        self._counter = 0

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        remaining_bytes = self._n - self._counter
        kept_fragment = raw_fragment[:remaining_bytes]
        deferred_fragment = raw_fragment[remaining_bytes:]
        self._counter += len(kept_fragment.data)
        remaining_bytes = self._n - self._counter
        return remaining_bytes == 0, kept_fragment, deferred_fragment, None


class Continuation(StopCondition):
    """
    Continuation stop-condition, used to stop reading when data has already been received
    and nothing has been received since then for the specified amount of time

    Parameters
    ----------
    continuation : float
    """

    def __init__(self, continuation: float) -> None:
        super().__init__()
        self.continuation = continuation
        self._last_fragment: float | None = None

    def __str__(self) -> str:
        return f"Continuation({self.continuation})"

    def __repr__(self) -> str:
        return self.__str__()

    def initiate_read(self, timestamp: float) -> None:
        self._last_fragment = timestamp

    def flush_read(self) -> None:
        self._last_fragment = None

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        deferred = Fragment(b"", raw_fragment.timestamp)
        kept = raw_fragment

        # if raw_fragment.timestamp is None:
        #    raise RuntimeError("Cannot evaluate fragment with no timestamp")
        # last_fragment can be none if no data was ever received
        if self._last_fragment is not None:
            continuation_timestamp = self._last_fragment + self.continuation
            stop = continuation_timestamp <= raw_fragment.timestamp
            next_event_timeout = continuation_timestamp
        else:
            stop = False
            next_event_timeout = None

        self._last_fragment = raw_fragment.timestamp

        return stop, kept, deferred, next_event_timeout

    def type(self) -> StopConditionType:
        return StopConditionType.CONTINUATION


class Total(StopCondition):
    """
    Total stop-condition, used to stop reading when data has already been received
    and the total read time exceeds the specified amount

    """

    def __init__(self, total: float) -> None:
        super().__init__()
        self.total = total
        self._start_time: float | None = None

    def __str__(self) -> str:
        return f"Total({self.total})"

    def __repr__(self) -> str:
        return self.__str__()

    def initiate_read(self, timestamp: float) -> None:
        self._start_time = timestamp

    def flush_read(self) -> None:
        self._start_time = None

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:
        kept = raw_fragment
        deferred = Fragment(b"", raw_fragment.timestamp)

        # if raw_fragment.timestamp is None:
        #     raise RuntimeError("Cannot evaluate fragment with no timestamp")

        if self._start_time is None:
            raise RuntimeError("Invalid start time")
        total_timestamp = self._start_time + self.total
        stop = total_timestamp <= raw_fragment.timestamp

        return stop, kept, deferred, total_timestamp

    def type(self) -> StopConditionType:
        return StopConditionType.TOTAL


class FragmentStopCondition(StopCondition):
    """
    Fragment stop-condition, used to stop on each piece of data received by the
    adapter

    """

    def __init__(self) -> None: ...

    def __str__(self) -> str:
        return "FragmentStopCondition()"

    def __repr__(self) -> str:
        return self.__str__()

    def initiate_read(self, timestamp: float) -> None:
        pass

    def flush_read(self) -> None:
        pass

    def evaluate(
        self, raw_fragment: Fragment
    ) -> tuple[bool, Fragment, Fragment, float | None]:

        return True, raw_fragment, Fragment(b"", raw_fragment.timestamp), None

    def type(self) -> StopConditionType:
        return StopConditionType.FRAGMENT


def termination_in_data(termination: bytes, data: bytes) -> tuple[int | None, int]:
    """
    Return the position (if it exists) and length of the termination (or part of it) inside data
    """
    p = None
    length = len(termination)
    # First check if the full termination is somewhere. If that's the case, data will be split
    try:
        p = data.index(termination)
        # If found, return that
    except ValueError:
        # If not, we'll try to find if part of the sequence is at the end, in that case
        # we'll return the length of the sequence that was found
        length -= 1
        while length > 0:
            if data[-length:] == termination[:length]:
                p = len(data) - length  # - 1
                break
            length -= 1

    return p, length



# File : serialport.py
# Author : Sébastien Deriaz
# License : GPL

"""
SerialPort module, allows communication with serial devices using
the OS layers (COMx, /dev/ttyUSBx or /dev/ttyACMx)

"""

import threading
from collections.abc import Callable
from dataclasses import dataclass
from enum import StrEnum
from types import EllipsisType

import serial
from serial.serialutil import PortNotOpenError

from syndesi.adapters.adapter_worker import AdapterEvent, HasFileno
from syndesi.component import Descriptor
from syndesi.tools.errors import AdapterOpenError, AdapterReadError
from syndesi.tools.types import NumberLike

from .adapter import Adapter
from .stop_conditions import Continuation, Fragment, StopCondition
from .timeout import Timeout


class Parity(StrEnum):
    """
    SerialPort parity setting, copied from pyserial
    """
    NONE = "N"
    EVEN = "E"
    ODD = "O"
    MARK = "M"
    SPACE = "S"


# pylint: disable=too-many-instance-attributes
@dataclass
class SerialPortDescriptor(Descriptor):
    """
    SerialPort descriptor that holds location (COMx or /dev/ttyx) and baudrate
    """

    DETECTION_PATTERN = r"(COM\d+|/dev[/\w\d]+):\d+"
    port: str
    baudrate: int | None = None
    bytesize: int = 8
    stopbits: int = 1
    parity: str = Parity.NONE.value
    rts_cts: bool = False
    dsr_dtr: bool = False
    xon_xoff: bool = False

    @staticmethod
    def from_string(string: str) -> "SerialPortDescriptor":
        parts = string.split(":")
        port = parts[0]
        baudrate = int(parts[1])
        return SerialPortDescriptor(port, baudrate)

    def set_default_baudrate(self, baudrate: int) -> bool:
        """
        Set the baudrate if it has not be defined before

        Parameters
        ----------
        baudrate : int
        """
        if self.baudrate is None:
            self.baudrate = baudrate
            return True

        return False

    def __str__(self) -> str:
        return f"{self.port}:{self.baudrate}"

    def is_initialized(self) -> bool:
        return self.baudrate is not None


class SerialPort(Adapter):
    """
    Serial communication adapter

    Parameters
    ----------
    port : str
        Serial port (COMx or ttyACMx)
    baudrate : int
        Baudrate
    """

    _open_ports: set[str] = set()
    _open_ports_lock = threading.Lock()

    def __init__(
        self,
        port: str,
        baudrate: int | None = None,
        *,
        timeout: Timeout | NumberLike | None | EllipsisType = ...,
        stop_conditions: StopCondition | list[StopCondition] | EllipsisType = ...,
        alias: str = "",
        bytesize: int = 8,
        stopbits: int = 1,
        parity: str = Parity.NONE.value,
        rts_cts: bool = False,
        xon_xoff: bool = False,
        dsr_dtr: bool = False,
        event_callback: Callable[[AdapterEvent], None] | None = None,
        auto_open: bool = True,
    ) -> None:
        """
        Instanciate new SerialPort adapter
        """
        self._port: serial.Serial | None = None
        descriptor = SerialPortDescriptor(
            port=port,
            baudrate=baudrate,
            bytesize=bytesize,
            stopbits=stopbits,
            parity=parity,
            rts_cts=rts_cts,
            dsr_dtr=dsr_dtr,
            xon_xoff=xon_xoff,
        )
        super().__init__(
            descriptor=descriptor,
            timeout=timeout,
            stop_conditions=stop_conditions,
            alias=alias,
            event_callback=event_callback,
            auto_open=auto_open,
        )
        self._descriptor: SerialPortDescriptor
        self._worker_descriptor: SerialPortDescriptor

        self._logger.info(
            f"Setting up SerialPort adapter {self._descriptor}, \
                timeout={timeout} and stop_conditions={self._stop_conditions}"
        )

        # self._bytesize = bytesize
        # self._stopbits = stopbits
        # self._parity = Parity(parity)
        # self._xonxoff = xon_xoff
        # self._dsrdtr = dsr_dtr

    def _default_timeout(self) -> Timeout:
        return Timeout(response=2, action="error")

    def _default_stop_conditions(self) -> list[StopCondition]:
        return [Continuation(0.1)]

    def _worker_open(self) -> None:
        super()._worker_open()
        self._worker_check_descriptor()

        if self._worker_descriptor.baudrate is None:
            raise AdapterOpenError(
                "Descriptor must be fully initialized to open the adapter"
            )

        if self._port is not None:
            self.close()

        port_name = self._worker_descriptor.port
        with self._open_ports_lock:
            if port_name in self._open_ports:
                raise AdapterOpenError(f"Port '{port_name}' is already in use")
            self._open_ports.add(port_name)

        try:
            self._port = serial.Serial(
                port=self._worker_descriptor.port,
                baudrate=self._worker_descriptor.baudrate,
                rtscts=self._worker_descriptor.rts_cts,
                bytesize=self._worker_descriptor.bytesize,
                parity=self._worker_descriptor.parity,
                stopbits=self._worker_descriptor.stopbits,
                xonxoff=self._worker_descriptor.xon_xoff,
                dsrdtr=self._worker_descriptor.dsr_dtr
            )
        except serial.SerialException as e:
            with self._open_ports_lock:
                self._open_ports.discard(port_name)
            if "No such file" in str(e):
                raise AdapterOpenError(
                    f"Port '{self._worker_descriptor.port}' was not found"
                ) from e
            raise AdapterOpenError("Unknown error") from e

        if self._port.isOpen():  # type: ignore
            self._logger.info(f"Adapter {self._worker_descriptor} opened")
        else:
            with self._open_ports_lock:
                self._open_ports.discard(port_name)
            self._logger.error(f"Failed to open adapter {self._worker_descriptor}")
            raise AdapterOpenError("Unknown error")

    def _worker_close(self) -> None:
        super()._worker_close()
        if self._port is not None:
            self._port.close()
            self._logger.info(f"Adapter {self._worker_descriptor} closed")
            self._port = None
            with self._open_ports_lock:
                self._open_ports.discard(self._worker_descriptor.port)

    async def aflush_read(self) -> None:
        await super().aflush_read()
        if self._port is not None:
            self._port.flush()

    def set_default_baudrate(self, baudrate: int) -> None:
        """
        Set baudrate

        Parameters
        ----------
        baudrate : int
        """
        if self._descriptor.set_default_baudrate(baudrate):
            self._update_descriptor()
            self.close()
            self.open()

    def _worker_write(self, data: bytes) -> None:
        super()._worker_write(data)
        if self._worker_descriptor.rts_cts:  # Experimental
            self._port.setRTS(True)  # type: ignore
        if self._port is not None:
            try:
                self._port.write(data)
            except (OSError, PortNotOpenError):
                pass

    def _worker_read(self, fragment_timestamp: float) -> Fragment:
        if self._port is None:
            raise AdapterReadError("Cannot read from non-initialized port")

        try:
            data = self._port.read_all()
        except (OSError, PortNotOpenError):
            data = None

        if data is None or data == b"":
            raise AdapterReadError(
                f"Error while reading from {self._worker_descriptor}"
            )

        return Fragment(data, fragment_timestamp)

    def _selectable(self) -> HasFileno | None:
        return self._port

    # def is_opened(self) -> bool:
    #     if self._port is not None:
    #         if self._port.isOpen():  # type: ignore
    #             return True

    #     return False



# File : shell_tools.py
# Author : Sébastien Deriaz
# License : GPL


# @dataclass
# class CommandSpec:
#     name: str
#     func: Callable[..., Any]  # the underlying function object
#     help: str = ""
#     aliases: tuple[str, ...] = ()
#     hidden: bool = False
#     dangerous: bool = False
#     owner_cls: type | None = None  # populated when collected
#     method_kind: str = "function"  # "instance" | "class" | "static" | "function"


# # ---- Module-level registry (optional)
# # If you like auto-discovery, classes decorated with @collect_cli_exports
# # get added here at import time.
# CLI_CLASS_REGISTRY: set[type] = set()


# def iter_all_cli_specs(classes: Iterable[type]) -> Iterable[CommandSpec]:
#     """Yield CommandSpec for the given classes (or CLI_CLASS_REGISTRY if you pass that)."""
#     for cls in classes:
#         for spec in getattr(cls, "__cli_exports__", ()):
#             yield from spec


# # ---- Decorators
# def cli_command(
#     name: str | None = None,
#     *,
#     help: str = "",
#     aliases: tuple[str, ...] = (),
#     hidden: bool = False,
#     dangerous: bool = False,
# ) -> None:
#     """
#     Decorate any callable (method or function) to mark it as CLI-visible.
#     Attach a CommandSpec to the function object. Class binding is assigned later.
#     """

#     def deco(fn: Callable[..., Any]) -> None:
#         # Note: we attach spec on the underlying function (works for instance funcs).
#         spec = CommandSpec(
#             name=(name or fn.__name__).replace("_", "-"),
#             func=fn,
#             help=help or (fn.__doc__ or "").strip(),
#             aliases=aliases,
#             hidden=hidden,
#             dangerous=dangerous,
#         )
#         fn.__cli_export__ = spec
#         return fn

#     return deco


# def collect_cli_exports(cls: type) -> None:
#     """
#     Class decorator: scan the class for decorated methods (instance, classmethod, staticmethod),
#     create a stable list of CommandSpec, and attach it as __cli_exports__.
#     Also registers the class in CLI_CLASS_REGISTRY for optional auto-discovery.
#     """
#     exports: list[CommandSpec] = []

#     for _, obj in cls.__dict__.items():
#         # instance method (function descriptor)
#         if inspect.isfunction(obj):
#             spec: CommandSpec | None = getattr(obj, "__cli_export__", None)
#             if spec:
#                 spec.owner_cls = cls
#                 spec.method_kind = "instance"
#                 exports.append(spec)
#             continue

#         # classmethod / staticmethod wrap the underlying function in .__func__
#         if isinstance(obj, classmethod):
#             fn = obj.__func__
#             spec = getattr(fn, "__cli_export__", None)
#             if spec:
#                 spec.owner_cls = cls
#                 spec.method_kind = "class"
#                 exports.append(spec)
#             continue

#         if isinstance(obj, staticmethod):
#             fn = obj.__func__
#             spec = getattr(fn, "__cli_export__", None)
#             if spec:
#                 spec.owner_cls = cls
#                 spec.method_kind = "static"
#                 exports.append(spec)
#             continue

#     cls.__cli_exports__ = tuple(exports)
#     CLI_CLASS_REGISTRY.add(cls)
#     return cls



# File : types.py
# Author : Sébastien Deriaz
# License : GPL
"""
Type tools
"""

from typing import TYPE_CHECKING, Any, TypeGuard

try:
    import numpy as np
except ImportError:
    np = None  # type: ignore[assignment]

if TYPE_CHECKING:
    import numpy as np_typing

    NumberLike = int | float | np_typing.number[Any]
else:
    NumberLike = int | float | np.number  # runtime will resolve string


def is_number(x: Any) -> TypeGuard[NumberLike]:
    """
    Check if the given X is an instance of int or float

    Parameters
    ----------
    X : any

    Returns
    -------
    result : bool
    """
    if np is None:
        return isinstance(x, int | float)
    return isinstance(x, int | float | np.number)


def assert_number(*args: Any) -> None:
    """
    Checks if the given argument(s) is a number.
    A TypeError is raised if it isn't the case

    Parameters
    ----------
    args
    """
    for arg in args:
        if not is_number(arg):
            raise TypeError(f"Variable {arg} should be a number")


def to_bytes(data: str | bytes) -> bytes:
    """
    Convert data to bytes array
    bytearray -> bytearray
    bytes -> bytes
    str -> bytes (UTF-8 encoding by default)
    """
    if isinstance(data, bytes):
        return data
    if isinstance(data, str):
        return data.encode("utf-8")
    raise ValueError(f"Invalid data type : {type(data)}")



# File : auto.py
# Author : Sébastien Deriaz
# License : GPL

"""
Automatic adapter function
This function is used to automatically choose an adapter based on the user's input
192.168.1.1 -> IP
COM4 -> Serial
/dev/tty* -> Serial
etc...
If an adapter class is supplied, it is passed through

Additionnaly, it is possible to do COM4:115200 so as to make the life of the user easier
Same with /dev/ttyACM0:115200
"""


import re

from syndesi.component import Descriptor

from .adapter import Adapter
from .ip import IP, IPDescriptor
from .serialport import SerialPort, SerialPortDescriptor
from .visa import Visa, VisaDescriptor

descriptors: list[type[Descriptor]] = [
    SerialPortDescriptor,
    IPDescriptor,
    VisaDescriptor,
]


def adapter_descriptor_by_string(string_descriptor: str) -> Descriptor:
    """
    Return a corresponding adapter descriptor from a string

    Parameters
    ----------
    string_descriptor : str

    Returns
    -------
    descriptor : Descriptor
    """
    for descriptor in descriptors:
        if re.match(descriptor.DETECTION_PATTERN, string_descriptor):
            x = descriptor.from_string(string_descriptor)
            return x
    raise ValueError(f"Could not parse descriptor string : {string_descriptor}")


def auto_adapter(adapter_or_string: Adapter | str) -> Adapter:
    """
    Create an adapter from a string or an adapter

    - <int>.<int>.<int>.<int>[:<int>] -> IP
    - x.y[:<int>] -> IP
    - COM<int> -> SerialPort
    - /dev/tty[ACM|USB]<int> -> SerialPort

    """
    if isinstance(adapter_or_string, Adapter):
        # Simply return it
        return adapter_or_string

    if isinstance(adapter_or_string, str):
        descriptor = adapter_descriptor_by_string(adapter_or_string)
        if isinstance(descriptor, IPDescriptor):
            return IP(
                address=descriptor.address,
                port=descriptor.port,
                transport=descriptor.transport.value,
            )
        if isinstance(descriptor, SerialPortDescriptor):
            return SerialPort(port=descriptor.port, baudrate=descriptor.baudrate)
        if isinstance(descriptor, VisaDescriptor):
            return Visa(descriptor=descriptor.descriptor)

        raise RuntimeError(f"Invalid descriptor : {descriptor}")

    raise ValueError(f"Invalid adapter : {adapter_or_string}")



# File : syndesi.py
# Author : Sébastien Deriaz
# License : GPL
"""
Main syndesi CLI script
"""

import argparse
from enum import Enum

from ..cli.shell import AdapterShell, AdapterType
from ..tools.logmanager import log
from ..version import __version__


class SyndesiCommands(Enum):
    """
    Syndesi script commands enum
    """

    SERIAL = "serial"
    IP = "ip"
    MODBUS = "modbus"
    VISA = "visa"


def main() -> None:
    """
    Main syndesi entry point
    """
    parser = argparse.ArgumentParser(
        prog="syndesi", description="Syndesi command line tool", epilog=""
    )

    parser.add_argument("--version", action="version", version=f"Syndesi {__version__}")
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument(
        "command",
        choices=[x.value for x in SyndesiCommands],
        help="Command, use syndesi <command> -h for help",
    )

    args, remaining_args = parser.parse_known_args()
    command = SyndesiCommands(args.command)

    if args.verbose:
        log("DEBUG", console=True)

    if command == SyndesiCommands.SERIAL:
        AdapterShell(AdapterType.SERIAL, remaining_args).run()
    elif command == SyndesiCommands.IP:
        AdapterShell(AdapterType.IP, remaining_args).run()
    elif command == SyndesiCommands.VISA:
        AdapterShell(AdapterType.VISA, remaining_args).run()
    else:
        raise NotImplementedError(f"Command '{command.value}' is not supported yet")


if __name__ == "__main__":
    main()



# File : component.py
# Author : Sébastien Deriaz
# License : GPL
"""
Component is the base of the main syndesi classes : Adapters, Protocols and Drivers
"""

import logging
from abc import ABC, abstractmethod
from concurrent.futures import Future
from dataclasses import dataclass, field
from enum import StrEnum
from types import EllipsisType
from typing import Generic, TypeVar

from syndesi.adapters.stop_conditions import Fragment, StopCondition, StopConditionType
from syndesi.adapters.timeout import Timeout
from syndesi.tools.errors import AdapterOpenError, WorkerThreadError

from .tools.log_settings import LoggerAlias


class Event:
    """Generic event, used to move information asynchronously from the adapter worker thread"""


class Descriptor(ABC):
    """
    Descriptor base class. A descriptor is a string to define the main parameters
    of an adapter (ip address, port, baudrate, etc...)
    """

    DETECTION_PATTERN = ""

    def __init__(self) -> None:
        return None

    @staticmethod
    @abstractmethod
    def from_string(string: str) -> "Descriptor":
        """
        Create a Descriptor class from a string
        """

    @abstractmethod
    def is_initialized(self) -> bool:
        """Return True if the descriptor is initialized"""


T = TypeVar("T")


@dataclass
class Frame(Generic[T]):
    """
    Adapter signal containing received data
    """

    stop_timestamp: float | None
    stop_condition_type: StopConditionType
    previous_read_buffer_used: bool
    response_delay: float | None

    @abstractmethod
    def get_payload(self) -> T:
        """
        Return frame payload
        """

    @abstractmethod
    def __str__(self) -> str: ...


@dataclass
class AdapterFrame(Frame[bytes]):
    """
    Adapter frame
    """

    fragments: list[Fragment] = field(default_factory=lambda: [])

    def get_payload(self) -> bytes:
        """
        Return all fragement data as a combined bytes array
        """
        return b"".join([f.data for f in self.fragments])

    def __str__(self) -> str:
        return f"AdapterFrame({self.get_payload()!r})"


R = TypeVar("R")


class ThreadCommand(Future[R]):
    """
    Command object completed by the worker thread.

    - .future is a concurrent.futures.Future => compatible with asyncio.wrap_future
    - .result() raises WorkerThreadError on command-timeout (worker not responding),
      not on device read timeouts (those are handled in the worker and surfaced as Adapter* errors).
    """

    def result(self, timeout: float | None = None) -> R:
        """
        Return the result of the thread command
        """
        try:
            return super().result(timeout=timeout)
        except TimeoutError:
            raise WorkerThreadError(
                f"No response from worker thread to {type(self).__name__} within {timeout}s"
            ) from None


class ReadScope(StrEnum):
    """
    Read scope

    NEXT : Only read data after the start of the read() call
    BUFFERED : Return any data that was present before the read() call
    """

    NEXT = "next"
    BUFFERED = "buffered"


class Component(ABC, Generic[T]):
    """Syndesi Component

    A Component is the elementary class of Syndesi. It is the base
    of all classes the user will be using
    """

    def __init__(self, logger_alias: LoggerAlias) -> None:
        #super().__init__()
        self._logger = logging.getLogger(logger_alias.value)

    # ==== open ====

    @abstractmethod
    def open(self) -> None:
        """Open the component"""

    @abstractmethod
    async def aopen(self) -> None:
        """Asynchronously open the component"""

    # ==== try_open ====

    async def atry_open(self) -> bool:
        """
        Async try to open communication with the device
        Return True if sucessful and False otherwise

        Returns
        -------
        success : bool
        """
        try:
            await self.aopen()
            return True
        except AdapterOpenError:
            return False

    def try_open(self) -> bool:
        """
        Try to open communication with the device
        Return True if sucessful and False otherwise

        Returns
        -------
        success : bool
        """
        try:
            self.open()
        except AdapterOpenError:
            return False
        return True

    # ==== close ====

    @abstractmethod
    def close(self) -> None:
        """Close the component"""

    @abstractmethod
    async def aclose(self) -> None:
        """Asynchronously close the component"""

    # ==== read_detailed ====

    @abstractmethod
    async def aread_detailed(
        self,
        timeout: Timeout | EllipsisType | None = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        scope: str = ReadScope.BUFFERED.value,
    ) -> Frame[T]:
        """Asynchronously read data from the component and return a Frame object"""

    @abstractmethod
    def read_detailed(
        self,
        timeout: Timeout | EllipsisType | None = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        scope: str = ReadScope.BUFFERED.value,
    ) -> Frame[T]:
        """Read data from the component and return a Frame object"""

    # ==== read ====

    @abstractmethod
    async def aread(
        self,
        timeout: Timeout | EllipsisType | None = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        scope: str = ReadScope.BUFFERED.value,
    ) -> T:
        """Asynchronously read data from the component"""

    @abstractmethod
    def read(
        self,
        timeout: Timeout | EllipsisType | None = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        scope: str = ReadScope.BUFFERED.value,
    ) -> T:
        """Read data from the component"""

    # ==== flush_read ====

    @abstractmethod
    async def aflush_read(self) -> None:
        """Clear input buffer"""

    @abstractmethod
    def flush_read(self) -> None:
        """Clear input buffer"""

    # ==== write ====

    @abstractmethod
    async def awrite(self, data: T) -> None:
        """Asynchronously write data to the component"""

    @abstractmethod
    def write(self, data: T) -> None:
        """Synchronously write data to the component"""

    # ==== query_detailed ====

    @abstractmethod
    async def aquery_detailed(
        self,
        payload: T,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        scope: str = ReadScope.BUFFERED.value,
    ) -> Frame[T]:
        """
        Asynchronously query the component and return a Frame object
        """

    @abstractmethod
    def query_detailed(
        self,
        payload: T,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        scope: str = ReadScope.BUFFERED.value,
    ) -> Frame[T]:
        """
        Synchronously query the component and return a Frame object
        """

    # ==== query ====

    async def aquery(
        self,
        payload: T,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        scope: str = ReadScope.BUFFERED.value,
    ) -> T:
        """Asynchronously query the component"""
        output_frame = await self.aquery_detailed(
            payload=payload,
            timeout=timeout,
            stop_conditions=stop_conditions,
            scope=scope,
        )
        return output_frame.get_payload()

    def query(
        self,
        payload: T,
        timeout: Timeout | None | EllipsisType = ...,
        stop_conditions: StopCondition | EllipsisType | list[StopCondition] = ...,
        scope: str = ReadScope.BUFFERED.value,
    ) -> T:
        """Query the component"""
        output_frame = self.query_detailed(
            payload=payload,
            timeout=timeout,
            stop_conditions=stop_conditions,
            scope=scope,
        )
        return output_frame.get_payload()

    # ==== Other ====

    @abstractmethod
    def is_open(self) -> bool:
        """Return True if the component is open"""



# File : logmanager.py
# Author : Sébastien Deriaz
# License : GPL
"""
Log manager implementation
"""
import logging
import threading
from typing import TextIO

from .log_settings import LoggerAlias


class LogManager:
    """
    Log manager, responsible for logging configuration, backend logging and
    printing to console and/or to a file
    """

    _lock = threading.Lock()

    DEFAULT_FORMATTER = logging.Formatter(
        "%(asctime)s:%(name)s:%(levelname)s:%(message)s"
    )
    DEFAULT_LOG_LEVEL = logging.ERROR

    def __init__(self) -> None:
        self._file_handler: logging.Handler | None = None
        self._all_loggers = False
        self._loggers: list[str] = []
        self._level = self.DEFAULT_LOG_LEVEL
        self._stream_handler: logging.StreamHandler[TextIO] | None = None

    def set_log_level(self, level: str | int) -> None:
        """
        Set log level
        """
        if isinstance(level, str):
            if not hasattr(logging, level.upper()):
                raise ValueError(f"Invalid log level: {level}")
            internal_level = getattr(logging, level.upper())
        elif isinstance(level, int):
            internal_level = level
        else:
            raise ValueError(f"Invalid level : {level}")

        self._level = internal_level
        self.update_loggers()

    def set_console_log(self, enabled: bool) -> None:
        """
        Enable or disable console logging
        """
        if enabled:
            self._stream_handler = logging.StreamHandler()
            self._stream_handler.setFormatter(self.DEFAULT_FORMATTER)
        else:
            self._stream_handler = None

    def set_log_file(self, file: str | None) -> None:
        """
        Define a file for loggers output
        """
        if file is not None:
            self._file_handler = logging.FileHandler(file)
            self._file_handler.setFormatter(self.DEFAULT_FORMATTER)
        else:
            if self._file_handler is not None:
                self._file_handler.close()
            self._file_handler = None

    def set_logger_filter(self, loggers: list[str] | str) -> None:
        """
        Select loggers
        """
        self._all_loggers = False
        self._loggers = []
        if isinstance(loggers, list):
            self._loggers = loggers
        elif isinstance(loggers, str):
            if loggers == "all":
                self._all_loggers = True
            else:
                self._loggers = [loggers]
        else:
            raise ValueError("Invalid argument loggers")

    def update_loggers(self) -> None:
        """
        Update configuration of each logger
        """
        # 1) Remove everything
        for alias in LoggerAlias:
            logger = logging.getLogger(alias.value)
            logger.handlers.clear()
        # 2) Update
        for alias in LoggerAlias:
            if self._all_loggers or alias.value in self._loggers:
                logger = logging.getLogger(alias.value)
                if self._file_handler is not None:
                    logger.addHandler(self._file_handler)
                if self._stream_handler is not None:
                    logger.addHandler(self._stream_handler)
                logger.setLevel(self._level)


log_manager = LogManager()


def log(
    level: str | int | None = None,
    console: bool | None = None,
    file: str | None = None,
    loggers: list[str] | str | None = None,
) -> None:
    """
    Configure syndesi logging

    Parameters
    ----------
        level : str or logging level
            . 'INFO'
            . 'CRITICAL'
            . 'ERROR'
            . 'WARNING'
            . 'INFO'
            . 'DEBUG'
        console : bool
            Print logging information to the console (True by default). Optional
        file : str
            File path, if None, file logging is disabled. Optionnal
        loggers : list
            Select which logger modules are updated (see LoggerAlias class). Optional
    """
    update = False
    if level is not None:
        log_manager.set_log_level(level)
        update = True
    if console is not None:
        log_manager.set_console_log(console)
        update = True
    if file is not None:
        log_manager.set_log_file(file)
        update = True
    if loggers is not None:
        log_manager.set_logger_filter(loggers)
        update = True

    if update:
        log_manager.update_loggers()



# # File : descriptors.py
# # Author : Sébastien Deriaz
# # License : GPL
# """
# Descriptors are classes that describe how an adapter is connected to its device.
# Depending on the protocol, they can hold strings, integers or enums
# """

# import re
# from abc import abstractmethod
# from dataclasses import dataclass
# from enum import Enum


# descriptors: list[type[Descriptor]] = [
#     SerialPortDescriptor,
#     IPDescriptor,
#     VisaDescriptor,
# ]


# def adapter_descriptor_by_string(string_descriptor: str) -> Descriptor:
#     """
#     Return a corresponding adapter descriptor from a string

#     Parameters
#     ----------
#     string_descriptor : str

#     Returns
#     -------
#     descriptor : Descriptor
#     """
#     for descriptor in descriptors:
#         if re.match(descriptor.DETECTION_PATTERN, string_descriptor):
#             x = descriptor.from_string(string_descriptor)
#             return x
#     raise ValueError(f"Could not parse descriptor string : {string_descriptor}")



# File : timeout.py
# Author : Sébastien Deriaz
# License : GPL
"""
This module holds the Timeout class, this class is meant for the user as a frontend for the
backend timeout management
"""

from enum import Enum
from types import EllipsisType
from typing import Any

from ..tools.types import NumberLike, is_number


class TimeoutAction(Enum):
    """
    Action on timeout expiration
    """
    ERROR = "error"
    RETURN_EMPTY = "return_empty"
    RETURN_NONE = "return_none"

class Timeout:
    """
    This class holds timeout information

    Parameters
    ----------
    response : float
        Time before the device responds
    action : str
        Action performed when a timeout occurs
        * ``error`` : raise a AdapterTimeoutError
        * ``return_empty`` : return b''
        * ``return_none`` : return None
    """

    DEFAULT_ACTION = TimeoutAction.ERROR

    def __init__(
        self,
        response: NumberLike | None | EllipsisType = ...,
        action: str | EllipsisType | TimeoutAction = ...,
    ) -> None:

        super().__init__()

        self._is_default_response = response is ...
        self._is_default_action = action is ...

        self.action: EllipsisType | TimeoutAction
        if action is ...:
            self.action = self.DEFAULT_ACTION
        else:
            self.action = TimeoutAction(action)

        self._response: EllipsisType | NumberLike | None = response

    def __str__(self) -> str:
        if self._response is ... or self._response is None:
            r = "..."
        else:
            r = f"{self._response:.3f}"
        if self.action is ...:
            a = "..."
        else:
            a = f"{self.action}"
        return f"Timeout({r}:{a})"

    def __repr__(self) -> str:
        return self.__str__()

    def set_default(self, default_timeout: "Timeout") -> None:
        """
        Set the default timeout (no effect if timeout is already set)

        Parameters
        ----------
        default_timeout : Timeout
        """
        if self._is_default_response:
            self._response = default_timeout.response()
        if self._is_default_action:
            self.action = default_timeout.action

    def response(self) -> NumberLike | None:
        """
        Return timeout response if it has been configured

        Returns
        -------
        response : NumberLike | None
        """
        if self._response is ...:
            return None
        if self._response is None:
            return None
        return self._response

    def is_initialized(self) -> bool:
        """
        Return True if the Timeout has been initialized, False otherwise
        """
        return self._response is not Ellipsis


def any_to_timeout(value: Any) -> Timeout:
    """
    Convert any input to a timeout (if possible)

    Parameters
    ----------
    value : None | NumberLike | Timeout

    Returns
    -------
    timeout : Timeout
    """
    if value is None:
        return Timeout(response=None)
    if is_number(value):
        return Timeout(response=float(value))
    if isinstance(value, Timeout):
        return value
    raise ValueError(f"Could not convert {value} to Timeout")



# File : terminal_tools.py
# Author : Sébastien Deriaz
# License : GPL


# class TerminalCompatible:
#     name = "..."
#     description = "..."
#     help = "..."
#     aliases = "..."
#     # prompt settings ? color ?

#     def handle_input(self, user_input: str | bytes) -> None:
#         pass
